#!/usr/bin/env python3
import copy
import logging
import os
import random
import sys
import uuid
import warnings
from logging import config
from multiprocessing.pool import AsyncResult
from typing import List, Optional, Union

import click
import yaml
import tqdm
from hyperopt import STATUS_OK, fmin, hp, tpe
from kube_fitness.tasks import IndividualDTO, TqdmToLogger
from yaml import Loader

from algorithms_for_tuning.utils import make_log_config_dict

ALG_ID = "bo"

SPACE = {
    "decor": hp.quniform("decor", 0, 1e5, 0.05),
    "n1": hp.quniform("n1", 0, 8, 1),
    "spb": hp.quniform("spb", 1e-3, 1e2, 0.05),
    "stb": hp.quniform("stb", 1e-3, 1e2, 0.05),
    "n2": hp.quniform("n2", 0, 8, 1),
    "sp1": hp.quniform("sp1", -1e2, -1e-3, 0.05),
    "st1": hp.quniform("st1", -1e2, -1e-3, 0.05),
    "n3": hp.quniform("n3", 0, 8, 1),
    "sp2": hp.quniform("sp2", -1e2, -1e-3, 0.05),
    "st2": hp.quniform("st2", -1e2, -1e-3, 0.05),
    "n4": hp.quniform("n4", 0, 8, 1),
    "B": hp.quniform("B", 0, 8, 1),
    "decor_2": hp.quniform("decor_2", 0, 1e5, 0.05),
}

warnings.filterwarnings("ignore")

logger = logging.getLogger("BO")

# getting config vars
if "FITNESS_CONFIG_PATH" in os.environ:
    filepath = os.environ["FITNESS_CONFIG_PATH"]
else:
    filepath = "../../algorithms_for_tuning/bo_algorithm/config.yaml"

with open(filepath, "r") as file:
    config = yaml.load(file, Loader=Loader)

    def estimate_fitness(
        population: List[IndividualDTO], _: bool = False, __: int = 2
    ) -> List[IndividualDTO]:
        results = []

        tqdm_out = TqdmToLogger(logger, level=logging.INFO)
        for p in tqdm(population, file=tqdm_out):
            individual = copy.deepcopy(p)
            individual.fitness_value = random.random()
            results.append(individual)

        return results

    def log_best_solution(
        individual: IndividualDTO,
        wait_for_result_timeout: Optional[float] = None,
        alg_args: Optional[str] = None,
    ) -> Union[IndividualDTO, AsyncResult]:
        ind = copy.deepcopy(individual)
        ind.fitness_value = random.random()
        return ind


NUM_FITNESS_EVALUATIONS = config["boAlgoParams"]["numEvals"]


class BigartmFitness:
    def __init__(self, dataset: str, exp_id: Optional[int] = None):
        self.dataset = dataset
        self.exp_id = exp_id
        # self.best_solution: Optional[IndividualDTO] = None

    def parse_kwargs(self, **kwargs):
        params = []
        params.append(kwargs.get("decor", 1))
        params.append(int(kwargs.get("n1", 1)))
        params.append(kwargs.get("spb", 1))
        params.append(kwargs.get("stb", 1))
        params.append(int(kwargs.get("n2", 1)))
        params.append(kwargs.get("sp1", 1))
        params.append(kwargs.get("st1", 1))
        params.append(int(kwargs.get("n3", 1)))
        params.append(kwargs.get("sp2", 1))
        params.append(kwargs.get("st2", 1))
        params.append(int(kwargs.get("n4", 1)))
        params.append(int(kwargs.get("B", 1)))
        params.append(kwargs.get("decor_2", 1))
        return params

    def make_individ(self, **kwargs):
        # TODO: adapt this function to work with baesyian optimization
        params = [float(i) for i in self.parse_kwargs(**kwargs)]
        params = params[:-1] + [0.0, 0.0, 0.0] + [params[-1]]
        return IndividualDTO(
            id=str(uuid.uuid4()),
            dataset=self.dataset,
            params=params,
            exp_id=self.exp_id,
            alg_id=ALG_ID,
        )

    def __call__(self, kwargs):
        population = [self.make_individ(**kwargs)]

        population = estimate_fitness(population)
        individ = population[0]

        # if self.best_solution is None or individ.fitness_value > self.best_solution.fitness_value:
        #     self.best_solution = copy.deepcopy(individ)

        return {"loss": -1 * individ.fitness_value, "status": STATUS_OK}


@click.command(context_settings=dict(allow_extra_args=True))
@click.option("--dataset", required=True, type=str, help="dataset name in the config")
@click.option(
    "--log-file",
    type=str,
    default="/var/log/tm-alg-bo.log",
    help="a log file to write logs of the algorithm execution to",
)
@click.option("--exp-id", required=True, type=int, help="mlflow experiment id")
def run_algorithm(dataset, log_file, exp_id):
    run_uid = uuid.uuid4() if not config["testMode"] else None
    logging_config = make_log_config_dict(filename=log_file, uid=run_uid)
    logging.config.dictConfig(logging_config)

    prepare_fitness_estimator()
    fitness = BigartmFitness(dataset, exp_id)
    best_params = fmin(
        fitness, SPACE, algo=tpe.suggest, max_evals=NUM_FITNESS_EVALUATIONS
    )
    best_solution = fitness.make_individ(**best_params)
    best_solution = log_best_solution(
        best_solution, wait_for_result_timeout=-1, alg_args=" ".join(sys.argv)
    )
    print(best_solution.fitness_value * -1)


if __name__ == "__main__":
    run_algorithm()



import logging
import os
import shutil
import uuid
from contextlib import contextmanager
from dataclasses import dataclass
from datetime import datetime
from typing import Optional, Tuple, List, cast, ContextManager, Dict

import artm
import gridfs
import mlflow
from pymongo import MongoClient

from autotm.utils import TimeMeasurements
from autotm.schemas import IndividualDTO, fitness_to_json
from autotm.fitness.tm import TopicModel


logger = logging.getLogger()


@dataclass
class TopicModelFiles:
    model_dir: str
    model_archive: str
    phi: str
    theta: str

    @property
    def files(self) -> Dict[str, str]:
        """
        :return: Dict of (name,filepath) pairs
        """
        return {"model": self.model_archive, "phi": self.phi, "theta": self.theta}


@contextmanager
def model_files(tm: TopicModel) -> ContextManager[TopicModelFiles]:
    uid = tm.uid
    base_path = f"/tmp/tm-{uid}"

    tmp_file = os.path.join(base_path, f"{uid}.artm")
    tmp_zip_file = os.path.join(base_path, f"{uid}.artm.zip")
    tmp_theta_file = os.path.join(base_path, f"theta-{uid}.csv")
    tmp_phi_file = os.path.join(base_path, f"phi-{uid}.csv")

    os.makedirs(base_path)

    tm.save_model(path=tmp_file)

    tmp_zip_file_base_name, _ = os.path.splitext(tmp_zip_file)
    shutil.make_archive(tmp_zip_file_base_name, "zip", tmp_file)

    model = cast(artm.ARTM, tm.model)
    theta_matrix = model.get_theta()
    phi_matrix = model.get_phi()

    theta_matrix.to_csv(tmp_theta_file, header=True)
    phi_matrix.to_csv(tmp_phi_file, header=True)

    yield TopicModelFiles(
        model_dir=tmp_file,
        model_archive=tmp_zip_file,
        phi=tmp_phi_file,
        theta=tmp_theta_file,
    )

    shutil.rmtree(base_path, ignore_errors=True)


def make_readable_topics(tm: TopicModel) -> str:
    topics = tm.get_topics()

    def topic_seq_num(pair: Tuple[str, List[str]]) -> int:
        tname, _ = pair
        if tname.startswith("main"):
            return int(tname[len("main") :])
        if tname.startswith("back"):
            return tm.topic_count + int(tname[len("back") :])
        return -1

    ordered_topics = sorted(topics.items(), key=topic_seq_num)
    topics_reprs = [
        f"{topic}:{os.linesep}{os.linesep.join(' '.join(words[i:i + 10]) for i in range(0, len(words), 10))}"
        for topic, words in ordered_topics
    ]
    full_repr = "\n\n".join(topics_reprs)
    return full_repr


def log_params_and_artifacts(
    tm: TopicModel,
    tm_files: TopicModelFiles,
    individual: IndividualDTO,
    time_metrics: TimeMeasurements,
    alg_args: Optional[str],
    is_tmp: bool = False
):
    logger.info("Logging params and artifacts to mlflow")
    logger.info(f"Created experiment_{individual.exp_id}")
    run_name = f"fitness-{individual.dataset}-{uuid.uuid4()}"
    if is_tmp:
        run_name+=f'_tmp_{individual.iteration_id}'

    try:
        experiment_id = mlflow.create_experiment(f"experiment_{individual.exp_id}")
    except:
        experiment = mlflow.get_experiment_by_name(f"experiment_{individual.exp_id}")
        experiment_id = experiment.experiment_id
        logger.info("Experiment exists, omitting creation")

    # mlflow.delete_experiment

    print(f"Experiment run name: {run_name}")
    # try:
    with mlflow.start_run(run_name=run_name, experiment_id=experiment_id):
        params = {
            "uid": tm.uid,
            "dataset": individual.dataset,
            "fitness_name": individual.fitness_name,
            "exp_id": individual.exp_id,
        }
        d = individual.make_params_dict()
        logger.debug(f"Params dict: {d}")
        params.update(d)

        artifact_path = "model.artm"
        theta_artifact_path = "theta.csv"
        phi_artifact_path = "phi.csv"
        individual_artifact_path = "individual.json"
        metrics_artifact_path = "metrics.json"
        time_metrics_artifact_path = "time_metrics.json"
        topics_artifact_path = "topics.json"
        readable_topics_artifact_path = "topics.txt"
        alg_args_artifact_path = "alg_args.txt"

        topics = tm.get_topics()
        full_repr_topics = make_readable_topics(tm)

        mlflow.log_params(params)
        mlflow.log_dict(individual.dict(), artifact_file=individual_artifact_path)
        mlflow.log_dict(individual.fitness_value, artifact_file=metrics_artifact_path)
        mlflow.log_dict(time_metrics, artifact_file=time_metrics_artifact_path)
        mlflow.log_artifact(local_path=tm_files.model_dir, artifact_path=artifact_path)
        mlflow.log_artifact(
            local_path=tm_files.theta, artifact_path=theta_artifact_path
        )
        mlflow.log_artifact(local_path=tm_files.phi, artifact_path=phi_artifact_path)
        mlflow.log_dict(topics, artifact_file=topics_artifact_path)
        mlflow.log_text(
            alg_args if alg_args else "", artifact_file=alg_args_artifact_path
        )
        mlflow.log_text(full_repr_topics, artifact_file=readable_topics_artifact_path)
    # except:
    #     logger.info("The model is already logged")

    logger.info("Logged params and artifacts to mlflow")


def log_stats(
    tm: TopicModel,
    tm_files: TopicModelFiles,
    individual: IndividualDTO,
    time_metrics: TimeMeasurements,
    alg_args: Optional[str],
):
    logger.info("Logging run stats to mongodb")

    uid = tm.uid

    if "MONGO_URI" not in os.environ:
        raise Exception("Unable to find mongo uri - MONGO_URI env variable not found")

    mongo_collection = os.environ.get("MONGO_COLLECTION", "tm_stats")

    with MongoClient(os.environ["MONGO_URI"]) as client:
        db = client.tm_experiments_runs
        logger.info("Writing main model's files to mongo GridFS")
        fs = gridfs.GridFS(db)

        gridfs_files = dict()
        for name, file_path in tm_files.files.items():
            with open(file_path, "rb") as f:
                file_id = fs.put(f, filename=f"{name}-{uid}.zip", parent_uid=f"{uid}")
            gridfs_files[f"{name}_file_id"] = file_id

        logger.info("Writing main stats")
        dt = datetime.now()
        stats = {
            "uid": str(uid),
            "datetime": dt.strftime("%Y-%m-%d %H:%M:%S.%f"),
            "timestamp": dt.timestamp(),
            "individual": individual.dict(),
            "alg_args": alg_args,
            "params": individual.make_params_dict(),
            "topics": tm.get_topics(),
            "time_metrics": time_metrics,
            **gridfs_files,
        }
        collection = db[mongo_collection]
        uid = collection.insert_one(stats).inserted_id

    logger.info(f"Inserted into mongodb stats, received uid {uid}")


@contextmanager
def succeed_or_log_error():
    try:
        yield
    except Exception as ex:
        logger.error(
            "Exception occured while trying to execute the action", exc_info=ex
        )



import logging
import os
import pickle
import time
import uuid
from collections import OrderedDict
from contextlib import contextmanager
from typing import Dict, Optional, Tuple, ContextManager, List
import multiprocessing as mp

import artm
import gensim.corpora as corpora
import numpy as np
import pandas as pd
from billiard.exceptions import SoftTimeLimitExceeded
from gensim.models.coherencemodel import CoherenceModel
from tqdm import tqdm

from autotm.fitness.external_scores import ts_bground, ts_uniform, ts_vacuous, switchp
from autotm.utils import (
    MetricsScores,
    AVG_COHERENCE_SCORE,
    TimeMeasurements,
    log_exec_timer,
)
from autotm.batch_vect_utils import SampleBatchVectorizer

logger = logging.getLogger()
logging.basicConfig(level="INFO")


class Dataset:
    _batches_path: str = "batches"
    _wv_path: str = "test_set_data_voc.txt"
    _cooc_dict_path: str = "cooc_dictionary.txt"
    _dictionary_path: str = "dictionary.txt"
    _vocab_path: str = "vocab.txt"
    _cooc_file_df_path: str = "cooc_df.txt"
    _cooc_file_tf_path: str = "cooc_tf.txt"
    _ppmi_dict_df_path: str = "ppmi_df.txt"
    _ppmi_dict_tf_path: str = "ppmi_tf.txt"
    _mutual_info_dict_path: str = "mutual_info_dict.pkl"
    _texts_path: str = "ppp.csv"
    _labels_path = "labels.pkl"

    def __init__(self, base_path: str, topic_count: int):
        self._base_path = base_path
        self._topic_count = topic_count

        self._labels = None
        self._dictionary: Optional[artm.Dictionary] = None
        self._batches: Optional[artm.BatchVectorizer] = None
        self._mutual_info_dict: Optional[Dict] = None

    @property
    def base_path(self) -> str:
        """
        :return: The path to the directory that contains this dataset files
        """
        return self._base_path

    @property
    def topic_count(self) -> int:
        return self._topic_count

    @property
    def labels(self):
        return self._labels

    @property
    def dictionary(self) -> artm.Dictionary:
        assert self._dictionary
        return self._dictionary

    @property
    def batches(self) -> artm.BatchVectorizer:
        assert self._batches
        return self._batches

    @property
    def mutual_info_dict(self) -> Dict:
        assert self._mutual_info_dict
        return self._mutual_info_dict

    @property
    def sample_batches(self) -> SampleBatchVectorizer:
        batches_dir_path = self._make_path(self._batches_path)
        return SampleBatchVectorizer(data_path=batches_dir_path, data_format="batches")

    @property
    def texts(self) -> List[str]:
        """
        :return: a list of PROCESSED texts of the corpus
        """
        texts_df_path = self._make_path(self._texts_path)
        df = pd.read_csv(texts_df_path)
        return df["processed_text"].tolist()

    @property
    def total_tokens(self) -> int:
        wv_path = self._make_path(self._wv_path)
        with open(wv_path) as f:
            total_tokens = sum(
                (
                    sum([int(elem.split(":")[1]) for elem in line.split()[1:]])
                    for line in f
                )
            )
        return total_tokens

    def verify_dataset_files(self):
        """
        Verifies that mandatory files exist in the dataset folder defined by :py:func:'.base_path'
        """
        paths = [
            self._make_path(self._dictionary_path),
            self._make_path(self._batches_path),
            self._make_path(self._mutual_info_dict_path),
            self._make_path(self._texts_path),
        ]

        for p in paths:
            assert os.path.exists(
                p
            ), f"Dataset file on path {p} should exist, but it doesn't"

    def load_dataset(self):
        """
        Partially loads this dataset files creating supplementary data_generator structures and entities like artm.BatchVectorizer
        """
        logger.info("Loading dataset entities")

        self.verify_dataset_files()

        start = time.time()
        labels_path = self._make_path(self._labels_path)
        dictionary_path = self._make_path(self._dictionary_path)
        batches_dir_path = self._make_path(self._batches_path)
        mutual_info_dict_path = self._make_path(self._mutual_info_dict_path)

        if os.path.exists(labels_path):
            logger.info(f"Reading labels {labels_path}")
            with open(labels_path, "rb") as f:
                self._labels = pickle.load(f)

        logger.info(f"Reading dictionary from {dictionary_path}")
        self._dictionary = artm.Dictionary()
        self._dictionary.load_text(dictionary_path)
        # TODO: tune the params
        self._dictionary.filter(min_df=5, max_tf=1000)

        self._batches = artm.BatchVectorizer(
            data_path=batches_dir_path, data_format="batches"
        )

        with open(mutual_info_dict_path, "rb") as handle:
            self._mutual_info_dict = pickle.load(handle)

        logging.info(f"Dataset entities initialization took {time.time() - start: .2f}")

    def _make_path(self, path):
        """
        Creates a full path out of a relative path or a filename
        :param path: a relative path or a filename
        :return: a full path
        """
        return os.path.join(self._base_path, path)


class TopicModelFactory:
    num_processors: Optional[int] = mp.cpu_count()
    experiments_path: str = "/tmp/tm_experiments"
    cached_dataset_settings: Dict[str, Dataset] = dict()

    @classmethod
    def init_factory_settings(
        cls,
        num_processors: Optional[int] = mp.cpu_count(),
        dataset_settings: Dict[str, Dict[str, object]] = None,
    ):
        cls.num_processors = num_processors
        cls.cached_dataset_settings = (
            {k: cls.init_dataset(v) for k, v in dataset_settings.items()}
            if dataset_settings
            else dict()
        )

    @classmethod
    def init_dataset(cls, settings) -> Dataset:
        assert "base_path" in settings and "topic_count" in settings
        dataset = Dataset(
            base_path=settings["base_path"], topic_count=settings["topic_count"]
        )
        dataset.load_dataset()
        return dataset

    def __init__(
        self,
        dataset_name: str,
        data_path: str,
        fitness_name: str,
        params: list,
        topic_count: Optional[int] = None,
        forced_update: bool = False,
        train_option: str = "offline",
    ):
        self.dataset_name = dataset_name
        self.data_path = data_path
        self.fitness_name = fitness_name
        self.params = params
        self.topic_count = topic_count
        self.forced_update = forced_update
        self.train_option = train_option
        self._custom_scores = []
        self.tm = None

    def __enter__(self) -> "TopicModel":
        # if self.dataset_name not in self.cached_dataset_settings:
        #     raise Exception(f"No settings for dataset {self.dataset_name}")

        # dataset = self.cached_dataset_settings[self.dataset_name]
        # t_count = self.topic_count if self.topic_count else dataset.topic_count
        #
        # logging.debug(f"Using the following settings: \n{dataset.base_path}")

        dataset = Dataset(base_path=self.data_path, topic_count=self.topic_count)
        dataset.load_dataset()

        uid = uuid.uuid4()

        if self.fitness_name == "default":
            logging.info(
                f"Using TM model: {TopicModel} according "
                f"to fitness name: {self.fitness_name}, topics count: {self.topic_count}"
            )
            self.tm = TopicModel(
                uid,
                self.experiments_path,
                self.topic_count,
                self.num_processors,
                dataset,
                self.params,
                train_option=self.train_option,
            )
        else:
            raise Exception(
                f"Unknown fitness name: {self.fitness_name}. Only the following ones are known: {['default']}"
            )

        self.tm.init_model()

        return self.tm

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.tm.dispose()


def type_check(res):
    res = list(res)
    for i in [1, 4, 7, 10, 11]:
        res[i] = int(res[i])
    return res


@contextmanager
def fit_tm_of_individual(
    dataset: str,
    data_path: str,
    params: list,
    fitness_name: str = "default",
    topic_count: Optional[int] = None,
    force_dataset_settings_checkout: bool = False,
    train_option: str = "offline",
) -> ContextManager[Tuple[TimeMeasurements, MetricsScores, "TopicModel"]]:
    params = type_check(params)

    start = time.time()

    with TopicModelFactory(
        dataset,
        data_path,
        fitness_name,
        params,
        topic_count,
        force_dataset_settings_checkout,
        train_option,
    ) as tm:
        try:
            with log_exec_timer("TM Training") as train_timer:
                tm.train()
            with log_exec_timer("Metrics calculation") as metrics_timer:
                fitness = tm.metrics_get_last_avg_vals(
                    texts=tm.dataset.texts, total_tokens=tm.dataset.total_tokens
                )

            time_metrics = {
                "train": train_timer.duration,
                "metrics": metrics_timer.duration,
            }
        except SoftTimeLimitExceeded as ex:
            raise ex
        except Exception:
            logger.warning(msg="Fitness calculation problem")
            fitness = {AVG_COHERENCE_SCORE: 0.0}
            time_metrics = {"train": -1, "metrics": -1}

        logging.info(f"Fitness estimation took {time.time() - start: .2f}")

        yield time_metrics, fitness, tm


class FitnessCalculatorWrapper:
    def __init__(self, dataset, data_path, topic_count, train_option):
        self.dataset = dataset
        self.data_path = data_path
        self.topic_count = topic_count
        self.train_option = train_option

    def run(self, params):
        print(params)
        params = list(params)
        params = params[:-1] + [0, 0, 0] + [params[-1]]
        fitness = calculate_fitness_of_individual(
            dataset=self.dataset,
            data_path=self.data_path,
            params=params,
            topic_count=self.topic_count,
            train_option=self.train_option,
        )
        result = fitness[AVG_COHERENCE_SCORE]
        print("Fitness: ", result)
        print()
        return -result


def calculate_fitness_of_individual(
    dataset: str,
    data_path: str,
    params: list,
    fitness_name: str = "default",
    topic_count: Optional[int] = None,
    force_dataset_settings_checkout: bool = False,
    train_option: str = "offline",
) -> MetricsScores:
    with fit_tm_of_individual(
        dataset,
        data_path,
        params,
        fitness_name,
        topic_count,
        force_dataset_settings_checkout,
        train_option,
    ) as result:
        time_metrics, fitness, tm = result

    return fitness


class TopicModel:
    def __init__(
        self,
        uid: uuid.UUID,
        experiments_path: str,
        topic_count: int,
        num_processors: int,
        dataset: Dataset,
        params: list,
        decor_test=False,
        train_option: str = "offline",
    ):
        self.uid = uid
        self.experiments_path: str = experiments_path
        self.topic_count: int = topic_count
        self.num_processors: int = num_processors
        self.dataset = dataset
        self.train_option = train_option

        self.model = None
        self.S = self.topic_count
        self.specific = ["main{}".format(i) for i in range(self.S)]

        self.save_path = os.path.join(self.experiments_path, "best_model")
        self.decor_test = decor_test

        if not os.path.exists(self.save_path):
            os.makedirs(self.save_path)

        self.__set_params(params)
        self.back = ["back{}".format(i) for i in range(self.B)]

    def init_model(self):
        self.model = artm.ARTM(
            num_topics=self.S + self.B,
            class_ids=["@default_class"],
            dictionary=self.dataset.dictionary,
            show_progress_bars=False,
            cache_theta=True,
            topic_names=self.specific + self.back,
            num_processors=self.num_processors,
        )

        self.__set_model_scores()

    def _early_stopping(self):
        coherences_main, coherences_back = self.__return_all_tokens_coherence(
            self.model, s=self.S, b=self.B
        )
        if len(coherences_main) < self.S or not any(coherences_main):
            return True
        return False

    # TODO: refactor option
    def train(self, option="online_v1"):
        if self.model is None:
            print("Initialise the model first!")
            return

        self.model.regularizers.add(
            artm.DecorrelatorPhiRegularizer(
                name="decorr", topic_names=self.specific, tau=self.decor
            )
        )
        self.model.regularizers.add(
            artm.DecorrelatorPhiRegularizer(
                name="decorr_2", topic_names=self.back, tau=self.decor_2
            )
        )
        if option == "offline":
            self.model.fit_offline(
                batch_vectorizer=self.dataset.batches, num_collection_passes=self.n1
            )
        elif option == "online_v1":
            self.model.fit_offline(
                batch_vectorizer=self.dataset.sample_batches,
                num_collection_passes=self.n1,
            )
        elif option == "online_v2":
            self.model.num_document_passes = self.n1
            self.model.fit_online(
                batch_vectorizer=self.dataset.sample_batches,
                update_every=self.num_processors,
            )

        if self.n1 > 0:
            if self._early_stopping():
                print("Early stopping is triggered")
                return

        #         if ((self.n2 != 0) and (self.B != 0)):
        if self.B != 0:
            self.model.regularizers.add(
                artm.SmoothSparseThetaRegularizer(
                    name="SmoothPhi", topic_names=self.back, tau=self.spb
                )
            )
            self.model.regularizers.add(
                artm.SmoothSparseThetaRegularizer(
                    name="SmoothTheta", topic_names=self.back, tau=self.stb
                )
            )
            if option == "offline":
                self.model.fit_offline(
                    batch_vectorizer=self.dataset.batches, num_collection_passes=self.n2
                )
            elif option == "online_v1":
                self.model.fit_offline(
                    batch_vectorizer=self.dataset.sample_batches,
                    num_collection_passes=self.n2,
                )
            elif option == "online_v2":
                self.model.num_document_passes = self.n2
                self.model.fit_online(
                    batch_vectorizer=self.dataset.sample_batches,
                    update_every=self.num_processors,
                )

        if self.n1 + self.n2 > 0:
            if self._early_stopping():
                print("Early stopping is triggered")
                return

        if self.n3 != 0:
            self.model.regularizers.add(
                artm.SmoothSparseThetaRegularizer(
                    name="SparsePhi", topic_names=self.specific, tau=self.sp1
                )
            )
            self.model.regularizers.add(
                artm.SmoothSparseThetaRegularizer(
                    name="SparseTheta", topic_names=self.specific, tau=self.st1
                )
            )
            if option == "offline":
                self.model.fit_offline(
                    batch_vectorizer=self.dataset.batches, num_collection_passes=self.n3
                )
            elif option == "online":
                self.model.fit_offline(
                    batch_vectorizer=self.dataset.sample_batches,
                    num_collection_passes=self.n3,
                )
            elif option == "online_v2":
                self.model.num_document_passes = self.n3
                self.model.fit_online(
                    batch_vectorizer=self.dataset.sample_batches,
                    update_every=self.num_processors,
                )

        if self.n1 + self.n2 + self.n3 > 0:
            if self._early_stopping():
                print("Early stopping is triggered")
                return

        if self.n4 != 0:
            self.model.regularizers["SparsePhi"].tau = self.sp2
            self.model.regularizers["SparseTheta"].tau = self.st2
            if option == "offline":
                self.model.fit_offline(
                    batch_vectorizer=self.dataset.batches, num_collection_passes=self.n4
                )
            elif option == "online_v1":
                self.model.fit_offline(
                    batch_vectorizer=self.dataset.sample_batches,
                    num_collection_passes=self.n4,
                )
            elif option == "online_v2":
                self.model.num_document_passes = self.n4
                self.model.fit_online(
                    batch_vectorizer=self.dataset.sample_batches,
                    update_every=self.num_processors,
                )

        if self.n1 + self.n2 + self.n3 > 0:
            if self._early_stopping():
                print("Early stopping is triggered")
                return

        print("Training is complete")

    def decor_train(self):
        if self.model is None:
            print("Initialise the model first")
            return

        self.model.regularizers.add(
            artm.DecorrelatorPhiRegularizer(
                name="decorr", topic_names=self.specific, tau=self.decor
            )
        )

    def dump_model(self, ii):
        self.model.dump_artm_model(os.path.join(self.save_path, "model_{}".format(ii)))

    def save_model(self, path):
        self.model.dump_artm_model(path)

    def print_topics(self):
        for i, (topic, top_tokens) in enumerate(self.get_topics().items()):
            print(topic)
            print(top_tokens)
            print()

    def get_topics(self):
        if "TopTokensScore" not in self.model.score_tracker:
            logger.warning(
                f"Key 'TopTokensScore' is not presented in the model's (uid={self.uid}) score_tracker. "
                f"Returning empty dict of topics."
            )
            return dict()
        res = self.model.score_tracker["TopTokensScore"].last_tokens
        topics = {topic: tokens[:50] for topic, tokens in res.items()}
        return topics

    def _get_avg_coherence_score(self, for_individ_fitness=False):
        coherences_main, coherences_back = self.__return_all_tokens_coherence(
            self.model, s=self.S, b=self.B
        )
        if for_individ_fitness:
            # print('COMPONENTS: ', np.mean(list(coherences_main.values())), np.min(list(coherences_main.values())))
            return np.mean(list(coherences_main.values())) + np.min(
                list(coherences_main.values())
            )
        return np.mean(list(coherences_main.values()))

    # added
    def _calculate_labels_coeff(self, topk=3):
        theta = self.model.get_theta()
        # theta.set_index('Unnamed: 0', inplace=True)
        documents_topics = OrderedDict()
        topics_documents = OrderedDict()
        for col in theta.columns:
            topics = theta.nlargest(topk, col).index.tolist()
            documents_topics[col] = topics
            for tp in topics:
                if tp not in topics_documents:
                    topics_documents[tp] = []
                topics_documents[tp].append(col)
        total_res = []
        for label in self.dataset.labels:
            docs = self.dataset.labels[label]
            all_lens = []
            for topic in topics_documents:
                all_lens.append(
                    len(set(topics_documents[topic]).intersection(set(docs)))
                )
            total_res.append(max(all_lens) / len(docs))
        coeff = np.mean(total_res)
        logging.info(f"Calculated coefficient: {coeff}")
        return coeff

    def dispose(self):
        """
        Disposes the model instance and removes the model from the object making it reusable again
        :return:
        """
        self.model.dispose()
        self.model = None

        log_files = [file for file in os.listdir(".") if file.startswith("bigartm.")]
        logging.info(f"Deleting bigartm logs: {log_files}")
        for file in log_files:
            os.remove(file)

    def __set_params(self, params_string):
        self.decor = params_string[0]
        self.n1 = params_string[1]

        if self.decor_test:
            return

        self.spb = params_string[2]
        self.stb = params_string[3]
        self.n2 = params_string[4]
        self.sp1 = params_string[5]
        self.st1 = params_string[6]
        self.n3 = params_string[7]
        self.sp2 = params_string[8]
        self.st2 = params_string[9]
        self.n4 = params_string[10]
        self.B = params_string[11]
        self.decor_2 = params_string[15]

    def __set_model_scores(self):
        self.model.scores.add(
            artm.PerplexityScore(
                name="PerplexityScore", dictionary=self.dataset.dictionary
            )
        )

        self.model.scores.add(
            artm.SparsityPhiScore(
                name="SparsityPhiScore",
                class_id="@default_class",
                topic_names=self.specific,
            )
        )
        self.model.scores.add(
            artm.SparsityThetaScore(
                name="SparsityThetaScore", topic_names=self.specific
            )
        )

        # Fraction of background words in the whole collection
        self.model.scores.add(
            artm.BackgroundTokensRatioScore(
                name="BackgroundTokensRatioScore", class_id="@default_class"
            )
        )

        # Kernel characteristics
        self.model.scores.add(
            artm.TopicKernelScore(
                name="TopicKernelScore",
                class_id="@default_class",
                topic_names=self.specific,
                probability_mass_threshold=0.5,
                dictionary=self.dataset.dictionary,
            )
        )

        # Looking at top tokens
        self.model.scores.add(
            artm.TopTokensScore(
                name="TopTokensScore", class_id="@default_class", num_tokens=100
            )
        )

    def __calculate_topic_coherence(self, tokens, top=50):
        tokens = tokens[:top]
        total_sum = 0
        for ix, token_1 in enumerate(tokens[:-1]):
            for ij, token_2 in enumerate(tokens[(ix + 1) :]):
                try:
                    total_sum += self.dataset.mutual_info_dict[
                        "{}_{}".format(token_1, token_2)
                    ]
                except KeyError:
                    total_sum += 0

        coherence = 2 / (top * (top - 1)) * total_sum
        return coherence

    def __return_all_tokens_coherence(self, model, s, b, top=50, return_backs=True):
        topics = list(model.score_tracker["TopTokensScore"].last_tokens.keys())

        res = model.score_tracker["TopTokensScore"].last_tokens

        topics_main = [topic for topic in topics if topic.startswith("main")]
        topics_back = [topic for topic in topics if topic.startswith("back")]

        all_topics_main = [i for i in range(s)]
        existing_topics_main = [int(i[4:]) for i in topics_main]
        inexisting_topics_main = [
            i for i in all_topics_main if i not in existing_topics_main
        ]

        all_topics_back = [i for i in range(b)]
        existing_topics_back = [int(i[4:]) for i in topics_back]
        inexisting_topics_back = [
            i for i in all_topics_back if i not in existing_topics_back
        ]

        coh_vals_main = {}
        # coherence for main topics
        for i, topic in tqdm(enumerate(topics_main)):
            coh_vals_main[topic] = self.__calculate_topic_coherence(
                res[topic][:50], top=top
            )
        for i, topic in tqdm(enumerate(inexisting_topics_main)):
            coh_vals_main["main{}".format(i)] = 0

        coh_vals_back = {}
        # coherence for back topics
        for i, topic in tqdm(enumerate(topics_back)):
            coh_vals_back[topic] = self.__calculate_topic_coherence(
                res[topic][:50], top=top
            )
        for i, topic in tqdm(enumerate(inexisting_topics_back)):
            coh_vals_back["back{}".format(i)] = 10  # penalty for not creating backs

        if return_backs:
            return coh_vals_main, coh_vals_back
        else:
            return coh_vals_main

    # TODO: fix
    def __return_all_coherence_types(
        self, model, S, only_specific=True, top=(10, 15, 20, 25, 30, 35, 40, 45, 50, 55)
    ):
        topics = list(model.score_tracker["TopTokensScore"].last_tokens.keys())
        coh_vals = {}

        res = model.score_tracker["TopTokensScore"].last_tokens
        if only_specific:
            topics = [topic for topic in topics if topic.startswith("main")]

        all_topics = [i for i in range(S)]
        existing_topics = [int(i[4:]) for i in topics]
        inexisting_topics = [i for i in all_topics if i not in existing_topics]

        for num_tokens in top:
            coh_vals["coherence_{}".format(num_tokens)] = {}
            for i, topic in tqdm(enumerate(topics)):
                coh_vals["coherence_{}".format(num_tokens)][
                    topic
                ] = self.__calculate_topic_coherence(res[topic][:100], top=num_tokens)
            for i, topic in tqdm(enumerate(inexisting_topics)):
                coh_vals["coherence_{}".format(num_tokens)][topic] = 0
        return coh_vals

    def metrics_get_avg_coherence_score(
        self, for_individ_fitness=False
    ) -> MetricsScores:
        coherences_main, coherences_back = self.__return_all_tokens_coherence(
            self.model, s=self.S, b=self.B
        )
        # commented after Masha's consultation
        # coeff = self._calculate_labels_coeff()
        coeff = 1.0
        if for_individ_fitness:
            print(
                "COMPONENTS: ",
                np.mean(list(coherences_main.values())),
                np.min(list(coherences_main.values())),
            )
            avg_coherence_score = (
                np.mean(list(coherences_main.values()))
                + np.min(list(coherences_main.values())) * coeff
            )
        else:
            avg_coherence_score = np.mean(list(coherences_main.values())) * coeff

        return {AVG_COHERENCE_SCORE: avg_coherence_score}

    # TODO: fix
    def metrics_get_last_avg_vals(
        self,
        texts,
        total_tokens,
        calculate_significance=False,
        calculate_npmi=False,
        calculate_switchp=False,
    ) -> MetricsScores:
        if calculate_significance:
            # turn off significance calculation
            phi_matrix = self.model.get_phi()
            theta_matrix = self.model.get_theta()

            phi_matrix = phi_matrix.to_numpy()
            theta_matrix = theta_matrix.to_numpy()
            topic_word_dist = phi_matrix[:, : self.S]  # getting subject scores
            doc_topic_dist = theta_matrix[: self.S, :]  #

            topic_significance_uni = np.mean(ts_uniform(topic_word_dist))
            topic_significance_vacuous = np.mean(
                ts_vacuous(doc_topic_dist, topic_word_dist, total_tokens)
            )
            topic_significance_back = np.mean(ts_bground(doc_topic_dist))
            print(
                f"Topic Significance - Uniform Distribution Over Words: {topic_significance_uni}"
            )
            print(
                f"Topic Significance - Vacuous Semantic Distribution: {topic_significance_vacuous}"
            )
            print(
                f"Topic Significance - Background Distribution: {topic_significance_back}"
            )
        else:
            topic_significance_uni = None
            topic_significance_vacuous = None
            topic_significance_back = None

        topic_names = list(
            self.model.score_tracker["TopTokensScore"].last_tokens.keys()
        )
        topics_dict = self.model.score_tracker["TopTokensScore"].last_tokens

        specific_topics = [topic for topic in topic_names if topic.startswith("main")]

        all_topics_flag = False
        if len(specific_topics) == self.S:
            print("Wow! all topics")
            all_topics_flag = True

        logger.info("Building dictionary")

        if calculate_npmi:
            splitted_texts = [text.split() for text in texts]
            id2word = corpora.Dictionary(splitted_texts)

            logger.info("Calculating NPMIs")

            npmis = (
                (
                    num_tokens,
                    calculate_npmi(splitted_texts, id2word, topics_dict, num_tokens),
                )
                for num_tokens in [15, 25, 50]
            )

            npmis = {
                k: v
                for num_tokens, npmi_list in npmis
                for k, v in [
                    (f"npmi_{num_tokens}", np.mean(npmi_list) if npmi_list else None),
                    (f"npmi_{num_tokens}_list", npmi_list),
                ]
            }

        else:
            npmis = {
                f"npmi_50": None,
                f"npmi_15": None,
                f"npmi_25": None,
                f"npmi_50_list": None,
            }

        if calculate_switchp:
            logger.info("Calculating switchp")

            if texts:
                switchp_scores = switchp(self.model.get_phi(), texts)
                switchp_scores = [
                    score if score is not None else 0 for score in switchp_scores
                ]
                print(f"SwitchP mean: {np.mean(switchp_scores)}")
            else:
                switchp_scores = None
        else:
            switchp_scores = [np.nan]

        background_ratio = self.model.score_tracker[
            "BackgroundTokensRatioScore"
        ].last_value
        control_perplexity = self.model.score_tracker["PerplexityScore"].last_value
        sparsity_phi = self.model.score_tracker["SparsityPhiScore"].last_value
        sparsity_theta = self.model.score_tracker["SparsityThetaScore"].last_value
        contrast = self.model.score_tracker["TopicKernelScore"].last_average_contrast
        purity = self.model.score_tracker["TopicKernelScore"].last_average_purity
        kernel_size = self.model.score_tracker["TopicKernelScore"].last_average_size

        avg_coherence_score = self.metrics_get_avg_coherence_score(
            for_individ_fitness=True
        )
        coherences = self.__return_all_coherence_types(
            self.model, self.S, only_specific=True
        )

        coherence_scores = dict()
        for i in range(10, 60, 5):
            coherence = list(coherences[f"coherence_{i}"].values())
            coherence_scores.update(
                {f"coherence_{i}_list": coherence, f"coherence_{i}": np.mean(coherence)}
            )

        scores_dict = {
            AVG_COHERENCE_SCORE: avg_coherence_score[AVG_COHERENCE_SCORE],
            "perplexityScore": control_perplexity,
            "backgroundTokensRatioScore": background_ratio,
            "contrast": contrast,
            "purity": purity,
            "kernelSize": kernel_size,
            "npmi_50_list": npmis["npmi_50_list"],  # npmi_values_50_list,
            "npmi_50": npmis["npmi_50"],
            "sparsity_phi": sparsity_phi,
            "sparsity_theta": sparsity_theta,
            "topic_significance_uni": topic_significance_uni,
            "topic_significance_vacuous": topic_significance_vacuous,
            "topic_significance_back": topic_significance_back,
            "switchP_list": switchp_scores,
            "switchP": np.nanmean(switchp_scores),
            "all_topics": all_topics_flag,
            **coherence_scores,
            **npmis,
        }

        return scores_dict



import os
import pandas as pd
import pymystem3
from nltk.corpus import stopwords, wordnet
from autotm.utils import parallelize_dataframe
import nltk

nltk.download("stopwords")
import re
from sklearn.feature_extraction.text import CountVectorizer

import spacy
from spacy.language import Language
from spacy_langdetect import LanguageDetector
from nltk.stem import WordNetLemmatizer

stop = stopwords.words("russian") + [" "] + stopwords.words("english")

r_html = re.compile(r"(\<[^>]*\>)")
r_punct = re.compile(r'[."\[\]/,()!?;:*#|\\%^$&{}~_`=-@]')
r_vk_ids = re.compile(r"(id{1}[0-9]*)")
r_num = re.compile(r"([0-9]+)")
r_white_space = re.compile(r"\s{2,}")
r_words = re.compile(r"\W+")
r_pat = re.compile(r"[aA-zZ]")

# check url pattern work
url_pattern = re.compile(
    r"http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\(\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+"
)

m = pymystem3.Mystem()
en_lemmatizer = WordNetLemmatizer()


@Language.factory("language_detector")
def language_detector(nlp, name):
    return LanguageDetector()


nlp_model = spacy.load("en_core_web_sm")
nlp_model.add_pipe("language_detector", last=True)


def new_html(text: str) -> str:
    """ This is my perfect function"""
    text = r_html.sub("", text)
    return text


def process_punkt(text: str) -> str:
    text = r_punct.sub(" ", text)
    text = r_vk_ids.sub(" ", text)
    text = r_num.sub(" ", text)
    text = r_white_space.sub(" ", text)
    return text.strip()


def get_lemma(word: str) -> str:
    lemma = wordnet.morphy(word)
    if lemma is None:
        return word
    else:
        return lemma


def tokens_num(text):
    return len(text.split(" "))


def lemmatize_text_ru(text: str) -> str:
    try:
        text = new_html(text)
    except:
        return ""
    text = text.lower()
    text = process_punkt(text)
    #         text = re.findall(r_rus, text)
    #         text = ' '.join(text)
    try:
        tokens = r_words.split(text)
    except:
        return ""
    tokens = (x for x in tokens if len(x) >= 3 and not x.isdigit())
    text = " ".join(tokens)
    tokens = m.lemmatize(text)
    tokens = (x for x in tokens if len(x) > 3)
    tokens = (x for x in tokens if x not in stop)
    tokens = (x for x in tokens if x.isalpha())
    text = " ".join(tokens)
    return text


def get_wordnet_pos(word):
    """Map POS tag to first character lemmatize() accepts"""
    tag = nltk.pos_tag([word])[0][1][0].lower()
    tag_dict = {
        "a": wordnet.ADJ,
        "n": wordnet.NOUN,
        "v": wordnet.VERB,
        "r": wordnet.ADV,
    }
    return tag_dict.get(tag, wordnet.NOUN)


def lemmatize_text_en(text):
    stop = stopwords.words("english")
    doc = nlp_model(text)
    detect_language = doc._.language
    if detect_language["language"] != "en":
        return " "
    text = text.lower()  # added
    text = process_punkt(text)
    try:
        # TODO: check this
        text_token = CountVectorizer().build_tokenizer()(text)
    except:
        return " "
    tokens = text.split()
    tokens = (x for x in tokens if len(x) >= 3 and r_pat.match(x) and not x.isdigit())
    text_rmstop = [i for i in tokens if i not in stop]

    text_stem = [
        en_lemmatizer.lemmatize(w, pos=get_wordnet_pos(w)) for w in text_rmstop
    ]
    return " ".join(text_stem)


def lemmatize_text(df, **kwargs):
    # print(kwargs)
    lang = kwargs["lang"]
    col_to_process = kwargs["col_to_process"]
    if lang == "ru":
        df["processed_text"] = df[col_to_process].apply(lemmatize_text_ru)
    elif lang == "en":
        df["processed_text"] = df[col_to_process].apply(lemmatize_text_en)
    else:
        print(f"The language {lang} is not known")
        raise NameError
    return df


def process_dataset(
    fname: str,
    col_to_process: str,
    save_path: str,
    lang: str = "ru",
    min_tokens_count: int = 3,
    n_cores: int = -1,
):
    """

    :param fname: Path to the dataset to process.
    :param col_to_process: The name of text column to be processed.
    :param save_path: Path where to store all the artifacts.
    :param lang: Language of the data (ru/en).
    :param min_tokens_count: Minimal amount of tokens to consider the further processing and topic modeling (3 by default).
    :param n_cores: Amount of cores for parallelization.
    :return:
    """
    os.makedirs(save_path, exist_ok=True)
    save_path = os.path.join(save_path, "ppp.csv")
    data = pd.read_csv(fname)
    data = parallelize_dataframe(
        data, lemmatize_text, n_cores, lang=lang, col_to_process=col_to_process
    )
    data["tokens_len"] = data["processed_text"].apply(tokens_num)
    data = data[data["tokens_len"] > min_tokens_count]
    data.to_csv(save_path, index=None)
    print("Saved to {}".format(save_path))



import copy
import gc
import logging
import math
import operator
import random
import sys
import time
import uuid
import warnings
from os.path import abspath, exists
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Union
import random

import numpy as np
from sklearn.svm import SVR
from sklearn.ensemble import BaggingRegressor
from sklearn.ensemble import RandomForestRegressor
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.gaussian_process.kernels import (
    RBF,
    Matern,
    WhiteKernel,
    ConstantKernel,
    ExpSineSquared,
    RationalQuadratic,
)
from sklearn.metrics import mean_squared_error
from sklearn.neural_network import MLPRegressor
from sklearn.tree import DecisionTreeRegressor

from autotm.algorithms_for_tuning.genetic_algorithm.mutation import mutation
from autotm.algorithms_for_tuning.genetic_algorithm.crossover import crossover
from autotm.algorithms_for_tuning.genetic_algorithm.selection import selection
from autotm.algorithms_for_tuning.individuals import make_individual, IndividualDTO
from autotm.algorithms_for_tuning.nelder_mead_optimization.nelder_mead import (
    NelderMeadOptimization,
)

from autotm.utils import AVG_COHERENCE_SCORE
from scipy.optimize import minimize
from autotm.visualization.dynamic_tracker import MetricsCollector

from autotm.fitness.tasks import estimate_fitness, log_best_solution

ALG_ID = "ga"
SPEEDUP = True

warnings.filterwarnings("ignore")
logger = logging.getLogger("GA_algo")


# TODO: Add fitness type
def set_surrogate_fitness(value, fitness_type="avg_coherence_score"):
    npmis = {
        f"npmi_50": None,
        f"npmi_15": None,
        f"npmi_25": None,
        f"npmi_50_list": None,
    }
    scores_dict = {
        fitness_type: value,
        "perplexityScore": None,
        "backgroundTokensRatioScore": None,
        "contrast": None,
        "purity": None,
        "kernelSize": None,
        "npmi_50_list": [None],  # npmi_values_50_list,
        "npmi_50": None,
        "sparsity_phi": None,
        "sparsity_theta": None,
        "topic_significance_uni": None,
        "topic_significance_vacuous": None,
        "topic_significance_back": None,
        "switchP_list": [None],
        "switchP": None,
        "all_topics": None,
        # **coherence_scores,
        **npmis,
    }
    return scores_dict


class Surrogate:
    def __init__(self, surrogate_name, **kwargs):
        self.name = surrogate_name
        self.kwargs = kwargs
        self.surrogate = None
        self.br_n_estimators = None
        self.br_n_jobs = None
        self.gpr_kernel = None

    def create(self):
        if self.name == "random-forest-regressor":
            self.surrogate = RandomForestRegressor(**self.kwargs)
        elif self.name == "mlp-regressor":
            if not self.br_n_estimators:
                self.br_n_estimators = self.kwargs["br_n_estimators"]
                del self.kwargs["br_n_estimators"]
                self.br_n_jobs = self.kwargs["n_jobs"]
                del self.kwargs["n_jobs"]
                self.kwargs["alpha"] = self.kwargs["mlp_alpha"]
                del self.kwargs["mlp_alpha"]
            self.surrogate = BaggingRegressor(
                base_estimator=MLPRegressor(**self.kwargs),
                n_estimators=self.br_n_estimators,
                n_jobs=self.br_n_jobs,
            )
        elif self.name == "GPR":  # tune ??
            if not self.gpr_kernel:
                kernel = self.kwargs["gpr_kernel"]
                del self.kwargs["gpr_kernel"]
                if kernel == "RBF":
                    self.gpr_kernel = 1.0 * RBF(1.0)
                elif kernel == "RBFwithConstant":
                    self.gpr_kernel = 1.0 * RBF(1.0) + ConstantKernel()
                elif kernel == "Matern":
                    self.gpr_kernel = 1.0 * Matern(1.0)
                elif kernel == "WhiteKernel":
                    self.gpr_kernel = 1.0 * WhiteKernel(1.0)
                elif kernel == "ExpSineSquared":
                    self.gpr_kernel = ExpSineSquared()
                elif kernel == "RationalQuadratic":
                    self.gpr_kernel = RationalQuadratic(1.0)
                self.kwargs["kernel"] = self.gpr_kernel
                self.kwargs["alpha"] = self.kwargs["gpr_alpha"]
                del self.kwargs["gpr_alpha"]
            self.surrogate = GaussianProcessRegressor(**self.kwargs)
        elif self.name == "decision-tree-regressor":
            try:
                if self.kwargs["max_depth"] == 0:
                    self.kwargs["max_depth"] = None
            except KeyError:
                logger.error("No max_depth")
            self.surrogate = DecisionTreeRegressor(**self.kwargs)
        elif self.name == "SVR":
            self.surrogate = SVR(**self.kwargs)
        # else:
        #     raise Exception('Undefined surr')

    def fit(self, X, y):
        logger.debug(f"X: {X}, y: {y}")
        self.create()
        self.surrogate.fit(X, y)

    def score(self, X, y):
        r_2 = self.surrogate.score(X, y)
        y_pred = self.surrogate.predict(X)
        mse = mean_squared_error(y, y_pred)
        rmse = np.sqrt(mse)
        return r_2, mse, rmse

    def predict(self, X):
        m = self.surrogate.predict(X)
        return m


def get_prediction_uncertanty(model, X, surrogate_name, percentile=90):
    interval_len = []
    if surrogate_name == "random-forest-regressor":
        for x in range(len(X)):
            preds = []
            for pred in model.estimators_:
                prediction = pred.predict(np.array(X[x]).reshape(1, -1))
                preds.append(prediction[0])
            err_down = np.percentile(preds, (100 - percentile) / 2.0)
            err_up = np.percentile(preds, 100 - (100 - percentile) / 2.0)
            interval_len.append(err_up - err_down)
    elif surrogate_name == "GPR":
        y_hat, y_sigma = model.predict(X, return_std=True)
        interval_len = list(y_sigma)
    elif surrogate_name == "decision-tree-regressor":
        raise NotImplementedError
    return interval_len


class ModelStorage:
    def __init__(self):
        self.stage_1_components = {}  # {config_id: id}
        self.stage_1_hyperp = {}  # {config_id: [[params1, params2]]}

    def model_search(self, model):
        raise NotImplementedError
        # for model.components


class GA:
    def __init__(
            self,
            dataset,
            data_path,
            num_individuals,
            num_iterations,
            mutation_type="mutation_one_param",
            crossover_type="blend_crossover",
            selection_type="fitness_prop",
            elem_cross_prob=0.2,
            num_fitness_evaluations: Optional[int] = 500,
            early_stopping_iterations: Optional[int] = 500,
            best_proc=0.3,
            alpha=None,
            exp_id: Optional[int] = None,
            surrogate_name=None,
            calc_scheme="type2",
            topic_count: Optional[int] = None,
            fitness_obj_type="single_objective",
            tag: Optional[str] = None,
            use_nelder_mead: bool = False,
            use_nelder_mead_in_mutation: bool = False,
            use_nelder_mead_in_crossover: bool = False,
            use_nelder_mead_in_selector: bool = False,
            train_option: str = "offline",
            **kwargs,
    ):
        """

        :param dataset: dataset name
        :param data_path: path to data
        :param num_individuals: number of individuals
        :param num_iterations: number of iterations
        :param mutation_type: type of mutation, available types ['mutation_one_param', 'combined', 'psm', 'positioning_mutation']
        :param crossover_type: type of crossover, available types ['crossover_pmx', 'crossover_one_point', 'blend_crossover']
        :param selection_type: type of selection, available types ['fitness_prop', 'rank_based']
        :param elem_cross_prob: probability of crossover
        :param num_fitness_evaluations: number of fitness evaluations in case of limited resources
        :param early_stopping_iterations: number of iterations when there is no significant changes in fitness to stop training
        :param best_proc: percentage of parents to be transferred to new generation
        :param alpha:
        :param exp_id:
        :param surrogate_name:
        :param calc_scheme: how to apply surrogates, available values: 'type 1' - to use surrogates through iteration, 'type 2' - calculating 70% on each iteration
        :param topic_count:
        :param fitness_obj_type:
        :param tag:
        :param use_nelder_mead: should Nelder-Mead enchan
        :param kwargs:
        """

        self.dataset = dataset

        if crossover_type == "blend_crossover":
            self.crossover_children = 1
        else:
            self.crossover_children = 2

        # if crossover_type == 'blend_crossover':
        #     self.crossover_children = 1
        # else:
        # self.crossover_children = 2
        self.data_path = data_path
        self.num_individuals = num_individuals
        self.num_iterations = num_iterations
        self.mutation = mutation(mutation_type)
        self.crossover = crossover(crossover_type)
        self.selection = selection(selection_type)
        self.elem_cross_prob = elem_cross_prob
        self.alpha = alpha
        self.evaluations_counter = 0
        self.num_fitness_evaluations = num_fitness_evaluations
        self.early_stopping_iterations = early_stopping_iterations
        self.fitness_obj_type = fitness_obj_type
        self.best_proc = best_proc
        self.all_params = []
        self.all_fitness = []
        if surrogate_name:
            self.surrogate = Surrogate(surrogate_name, **kwargs)
        else:
            self.surrogate = None
        self.exp_id = exp_id
        self.calc_scheme = calc_scheme
        self.topic_count = topic_count
        self.tag = tag
        # hyperparams
        self.set_regularizer_limits()
        self.use_nelder_mead = use_nelder_mead
        self.use_nelder_mead_in_mutation = use_nelder_mead_in_mutation
        self.use_nelder_mead_in_crossover = use_nelder_mead_in_crossover
        self.use_nelder_mead_in_selectior = use_nelder_mead_in_selector
        self.train_option = train_option
        self.metric_collector = MetricsCollector(
            dataset=self.dataset, n_specific_topics=topic_count
        )
        self.crossover_changes_dict = (
            {}
        )  # generation, parent_1_params, parent_2_params, ...

    def set_regularizer_limits(
            self,
            low_decor=0,
            high_decor=1e5,
            low_n=0,
            high_n=30,
            low_back=0,
            high_back=5,
            low_spb=0,
            high_spb=1e2,
            low_spm=-1e-3,
            high_spm=1e2,
            low_sp_phi=-1e3,
            high_sp_phi=1e3,
            low_prob=0,
            high_prob=1,
    ):
        self.high_decor = high_decor
        self.low_decor = low_decor
        self.low_n = low_n
        self.high_n = high_n
        self.low_back = low_back
        self.high_back = high_back
        self.high_spb = high_spb
        self.low_spb = low_spb
        self.low_spm = low_spm
        self.high_spm = high_spm
        self.low_sp_phi = low_sp_phi
        self.high_sp_phi = high_sp_phi
        self.low_prob = low_prob
        self.high_prob = high_prob

    def init_individ(self, base_model=False):
        val_decor = np.random.uniform(low=self.low_decor, high=self.high_decor, size=1)[
            0
        ]
        var_n = np.random.randint(low=self.low_n, high=self.high_n, size=4)
        var_back = np.random.randint(low=self.low_back, high=self.high_back, size=1)[0]
        var_sm = np.random.uniform(low=self.low_spb, high=self.high_spb, size=2)
        var_sp = np.random.uniform(low=self.low_sp_phi, high=self.high_sp_phi, size=4)
        ext_mutation_prob = np.random.uniform(
            low=self.low_prob, high=self.high_prob, size=1
        )[0]
        ext_elem_mutation_prob = np.random.uniform(
            low=self.low_prob, high=self.high_prob, size=1
        )[0]
        ext_mutation_selector = np.random.uniform(
            low=self.low_prob, high=self.high_prob, size=1
        )[0]
        val_decor_2 = np.random.uniform(
            low=self.low_decor, high=self.high_decor, size=1
        )[0]
        params = [
            val_decor,
            var_n[0],
            var_sm[0],
            var_sm[1],
            var_n[1],
            var_sp[0],
            var_sp[1],
            var_n[2],
            var_sp[2],
            var_sp[3],
            var_n[3],
            var_back,
            ext_mutation_prob,
            ext_elem_mutation_prob,
            ext_mutation_selector,
            val_decor_2,
        ]
        if base_model:
            for i in [0, 4, 7, 10, 11, 15]:
                params[i] = 0
        params = [float(i) for i in params]
        return params

    def init_population(self):
        list_of_individuals = []
        for i in range(self.num_individuals):
            if i == 0:
                dto = IndividualDTO(
                    id=str(uuid.uuid4()),
                    data_path=self.data_path,
                    dataset=self.dataset,
                    params=self.init_individ(base_model=True),
                    exp_id=self.exp_id,
                    alg_id=ALG_ID,
                    iteration_id=0,
                    topic_count=self.topic_count,
                    tag=self.tag,
                    train_option=self.train_option,
                )
            else:
                dto = IndividualDTO(
                    id=str(uuid.uuid4()),
                    data_path=self.data_path,
                    dataset=self.dataset,
                    params=self.init_individ(),
                    exp_id=self.exp_id,
                    alg_id=ALG_ID,
                    iteration_id=0,
                    topic_count=self.topic_count,
                    tag=self.tag,
                    train_option=self.train_option,
                )
            # TODO: improve heuristic on search space
            list_of_individuals.append(make_individual(dto=dto))
        population_with_fitness = estimate_fitness(list_of_individuals)

        self.save_params(population_with_fitness)
        if self.surrogate is not None and self.calc_scheme == "type2":
            self.surrogate.fit(np.array(self.all_params), np.array(self.all_fitness))
            logger.info("Surrogate is initialized!")
        return population_with_fitness

    def _calculate_uncertain_res(self, generation, iteration_num: int, proc=0.3):
        X = np.array([individ.dto.params for individ in generation])
        certanty = get_prediction_uncertanty(
            self.surrogate.surrogate, X, self.surrogate.name
        )
        recalculate_num = int(np.floor(len(certanty) * proc))
        logger.info(f"Certanty values: {certanty}")

        certanty, X = (
            list(t) for t in zip(*sorted(zip(certanty, X.tolist()), reverse=True))
        )  # check
        calculated = []
        for params in X[:recalculate_num]:
            dto = IndividualDTO(
                id=str(uuid.uuid4()),
                data_path=self.data_path,
                params=[float(i) for i in params],
                dataset=self.dataset,
                exp_id=self.exp_id,
                alg_id=ALG_ID,
                iteration_id=iteration_num,
                topic_count=self.topic_count,
                tag=self.tag,
                train_option=self.train_option,
            )
            calculated.append(make_individual(dto=dto))

        calculated = estimate_fitness(calculated)

        self.all_params += [individ.dto.params for individ in calculated]
        self.all_fitness += [
            individ.dto.fitness_value["avg_coherence_score"] for individ in calculated
        ]

        pred_y = self.surrogate.predict(X[recalculate_num:])
        for ix, params in enumerate(X[recalculate_num:]):
            dto = IndividualDTO(
                id=str(uuid.uuid4()),
                data_path=self.data_path,
                params=params,
                dataset=self.dataset,
                fitness_value=set_surrogate_fitness(pred_y[ix]),
                exp_id=self.exp_id,
                alg_id=ALG_ID,
                topic_count=self.topic_count,
                tag=self.tag,
                train_option=self.train_option,
            )
            calculated.append(make_individual(dto=dto))
        return calculated

    def save_params(self, population):
        params_and_f = [
            (copy.deepcopy(individ.params), individ.fitness_value)
            for individ in population
            if individ.fitness_value not in self.all_fitness
        ]

        def check_val(fval):
            return not (fval is None or math.isnan(fval) or math.isinf(fval))

        def check_params(p):
            return all(check_val(el) for el in p)

        clean_params_and_f = []
        for p, f in params_and_f:
            if not check_params(p) or not check_val(f):
                logger.warning(
                    f"Bad params or fitness found. Fitness: {f}. Params: {p}."
                )
            else:
                clean_params_and_f.append((p, f))

        pops = [p for p, _ in clean_params_and_f]
        fs = [f for _, f in clean_params_and_f]

        self.all_params += pops
        self.all_fitness += fs

    def surrogate_calculation(self, population):
        X_val = np.array([copy.deepcopy(individ.params) for individ in population])
        y_pred = self.surrogate.predict(X_val)
        if not SPEEDUP:
            y_val = np.array([individ.fitness_value for individ in population])

            def check_val(fval):
                return not (fval is None or math.isnan(fval) or math.isinf(fval))

            def check_params(p):
                return all(check_val(el) for el in p)

            clean_params_and_f = []
            for i in range(len(y_val)):
                if not check_params(X_val[i]) or not check_val(y_val[i]):
                    logger.warning(
                        f"Bad params or fitness found. Fitness: {y_val[i]}. Params: {X_val[i]}."
                    )
                else:
                    clean_params_and_f.append((X_val[i], y_val[i]))

            X_val = clean_params_and_f[0]
            y_val = clean_params_and_f[1]
            r_2, mse, rmse = self.surrogate.score(X_val, y_val)
            logger.info(f"Real values: {list(y_val)}")
            logger.info(f"Predicted values: {list(y_pred)}")
            logger.info(f"R^2: {r_2}, MSE: {mse}, RMSE: {rmse}")
        for ix, individ in enumerate(population):
            individ.dto.fitness_value = set_surrogate_fitness(y_pred[ix])
        return population

    def _check_param(self, param, bounds):
        param = min(param, bounds[1])
        param = max(param, bounds[0])
        return param

    def check_params_bounds(self, params):
        # smooth
        for i in [2, 3]:
            params[i] = self._check_param(params[i], (self.low_spb, self.high_spb))
        # sparce
        for i in [5, 6, 8, 9]:
            params[i] = self._check_param(
                params[i], (self.low_sp_phi, self.high_sp_phi)
            )
        # iterations
        for i in [1, 4, 7, 10]:
            params[i] = float(int(params[i]))
            params[i] = self._check_param(params[i], (self.low_n, self.high_n))
        # back
        for i in [11]:
            params[i] = float(int(params[i]))
            params[i] = self._check_param(params[i], (self.low_back, self.high_back))
        # mutation
        for i in [12, 13, 14]:
            params[i] = self._check_param(params[i], (self.low_prob, self.high_prob))
        # decorrelation
        for i in [0, 15]:
            params[i] = self._check_param(params[i], (self.low_decor, self.high_decor))
        return params

    def run_crossover(self, pairs_generator, surrogate_iteration, iteration_num: int):
        new_generation = []

        crossover_changes = {
            "parent_1_params": [],
            "parent_2_params": [],
            "parent_1_fitness": [],
            "parent_2_fitness": [],
            "child_id": [],
        }

        for i, j in pairs_generator:
            if i is None:
                break

            parent_1 = copy.deepcopy(i.params)
            parent_2 = copy.deepcopy(j.params)

            if self.crossover_children == 2:
                child_1, child_2 = self.crossover(
                    parent_1=parent_1,
                    parent_2=parent_2,
                    elem_cross_prob=self.elem_cross_prob,
                    alpha=self.alpha,
                )

                child_1 = self.check_params_bounds(child_1)
                child_2 = self.check_params_bounds(child_2)

                child1_dto = IndividualDTO(
                    id=str(uuid.uuid4()),
                    data_path=self.data_path,
                    dataset=self.dataset,
                    params=child_1,
                    exp_id=self.exp_id,
                    alg_id=ALG_ID,
                    iteration_id=iteration_num,
                    topic_count=self.topic_count,
                    tag=self.tag,
                    train_option=self.train_option,
                )
                child2_dto = IndividualDTO(
                    id=str(uuid.uuid4()),
                    data_path=self.data_path,
                    dataset=self.dataset,
                    params=child_2,
                    exp_id=self.exp_id,
                    alg_id=ALG_ID,
                    iteration_id=iteration_num,
                    topic_count=self.topic_count,
                    tag=self.tag,
                    train_option=self.train_option,
                )

                new_generation.append(make_individual(dto=child1_dto))
                new_generation.append(make_individual(dto=child2_dto))
                self.evaluations_counter += 2
            else:
                child_1 = self.crossover(
                    parent_1=parent_1,
                    parent_2=parent_2,
                    elem_cross_prob=self.elem_cross_prob,
                    alpha=self.alpha,
                )

                child_1 = self.check_params_bounds(child_1)

                child1_dto = IndividualDTO(
                    id=str(uuid.uuid4()),
                    data_path=self.data_path,
                    dataset=self.dataset,
                    params=child_1,
                    exp_id=self.exp_id,
                    alg_id=ALG_ID,
                    iteration_id=iteration_num,
                    topic_count=self.topic_count,
                    tag=self.tag,
                    train_option=self.train_option,
                )
                new_generation.append(make_individual(dto=child1_dto))

                self.evaluations_counter += 1

            crossover_changes["parent_1_params"].append(i.params)
            crossover_changes["parent_2_params"].append(j.params)
            crossover_changes["parent_1_fitness"].append(i.fitness_value)
            crossover_changes["parent_2_fitness"].append(j.fitness_value)
            crossover_changes["child_id"].append(len(new_generation) - 1)

        logger.info(f"CURRENT COUNTER: {self.evaluations_counter}")

        if len(new_generation) > 0:
            fitness_calc_time_start = time.time()
            if not SPEEDUP or not self.surrogate:
                new_generation = estimate_fitness(
                    new_generation
                )  # TODO: check the order
                self.save_params(new_generation)
            logger.info(f"ize of the new generation is {len(new_generation)}")
            logger.info(
                f"TIME OF THE FITNESS FUNCTION IN CROSSOVER: {time.time() - fitness_calc_time_start}"
            )

            for i in range(len(crossover_changes["parent_1_params"])):
                self.metric_collector.save_crossover(
                    generation=iteration_num,
                    parent_1=crossover_changes["parent_1_params"][i],
                    parent_2=crossover_changes["parent_2_params"][i],
                    parent_1_fitness=crossover_changes["parent_1_fitness"][i],
                    parent_2_fitness=crossover_changes["parent_2_fitness"][i],
                    child_1=new_generation[crossover_changes["child_id"][i]].params,
                    child_1_fitness=new_generation[
                        crossover_changes["child_id"][i]
                    ].fitness_value,
                )

            if self.surrogate:
                if self.calc_scheme == "type1":
                    if surrogate_iteration:
                        self.surrogate_calculation(new_generation)
                    elif not surrogate_iteration and SPEEDUP:
                        new_generation = estimate_fitness(new_generation)
                        self.save_params(new_generation)
                elif self.calc_scheme == "type2":
                    new_generation = self._calculate_uncertain_res(
                        new_generation, iteration_num=iteration_num
                    )
                    self.save_params(new_generation)

        return new_generation, crossover_changes

    def apply_nelder_mead(self, starting_points_set, num_gen, num_iterations=2):
        nelder_opt = NelderMeadOptimization(
            data_path=self.data_path,
            dataset=self.dataset,
            exp_id=self.exp_id,
            topic_count=self.topic_count,
            train_option=self.train_option,
        )
        new_population = []
        for point in starting_points_set:
            st_point = point[:12] + [point[15]]
            res = nelder_opt.run_algorithm(
                num_iterations=num_iterations, ini_point=st_point
            )
            solution = list(res["x"])
            solution = (
                    solution[:-1] + point[12:15] + [solution[-1]]
            )  # TODO: check mutation ids
            fitness = -res.fun
            solution_dto = IndividualDTO(
                id=str(uuid.uuid4()),
                data_path=self.data_path,
                dataset=self.dataset,
                params=solution,
                exp_id=self.exp_id,
                alg_id=ALG_ID,
                iteration_id=num_gen,
                topic_count=self.topic_count,
                tag=self.tag,
                fitness_value={AVG_COHERENCE_SCORE: fitness},
                train_option=self.train_option,
            )

            new_population.append(make_individual(dto=solution_dto))
        return new_population

    def run(self, verbose=False):
        self.evaluations_counter = 0
        ftime = str(int(time.time()))

        # os.makedirs(LOG_FILE_PATH, exist_ok=True)

        logger.info(f"Starting experiment: {ftime}")

        logger.info(
            f"ALGORITHM PARAMS  number of individuals {self.num_individuals}; "
            f"number of fitness evals "
            f"{self.num_fitness_evaluations if self.num_fitness_evaluations else 'unlimited'}; "
            f"number of early stopping iterations "
            f"{self.early_stopping_iterations if self.early_stopping_iterations else 'unlimited'}; "
            f"crossover prob {self.elem_cross_prob}"
        )

        # population initialization
        population = self.init_population()

        self.evaluations_counter = self.num_individuals

        logger.info("POPULATION IS CREATED")

        x, y = [], []
        high_fitness = 0
        surrogate_iteration = False
        best_val_so_far = -10
        early_stopping_counter = 0

        run_id = str(uuid.uuid4())

        for ii in range(self.num_iterations):
            iteration_start_time = time.time()
            before_mutation = []  # individual
            id_mutation = []  # new individual id

            logger.info(f"ENTERING GENERATION {ii}")

            if self.surrogate is not None:
                surrogate_iteration = False
                if ii % 2 != 0:
                    surrogate_iteration = True

            population.sort(key=operator.attrgetter("fitness_value"), reverse=True)
            pairs_generator = self.selection(
                population=population,
                best_proc=self.best_proc,
                children_num=self.crossover_children,
            )

            logger.info(f"PAIRS ARE CREATED")

            # Crossover
            new_generation, crossover_changes = self.run_crossover(
                pairs_generator, surrogate_iteration, iteration_num=ii
            )

            new_generation.sort(key=operator.attrgetter("fitness_value"), reverse=True)
            population.sort(key=operator.attrgetter("fitness_value"), reverse=True)

            logger.info("CROSSOVER IS OVER")

            if self.use_nelder_mead_in_crossover:
                # TODO: implement Nelder-Mead here
                pass

            if (
                    self.num_fitness_evaluations
                    and self.evaluations_counter >= self.num_fitness_evaluations
            ):
                bparams = "".join([str(i) for i in population[0].params])
                self.metric_collector.save_fitness(
                    generation=ii,
                    params=[i.params for i in population],
                    fitness=[i.fitness_value for i in population],
                )
                logger.info(
                    f"TERMINATION IS TRIGGERED: EVAL NUM."
                    f"DATASET {self.dataset}."
                    f"TOPICS NUM {self.topic_count}."
                    f"RUN ID {run_id}."
                    f"THE BEST FITNESS {population[0].fitness_value}."
                    f"THE BEST PARAMS {bparams}."
                    f"ITERATION TIME {time.time() - iteration_start_time}."
                )
                # return population[0].fitness_value
                break

            del pairs_generator
            gc.collect()

            # population_params = [copy.deepcopy(individ.params) for individ in population]

            the_best_guy_params = copy.deepcopy(population[0].params)
            new_generation = [
                individ
                for individ in new_generation
                if individ.params != the_best_guy_params
            ]

            new_generation_n = min(
                (
                        self.num_individuals
                        - int(np.ceil(self.num_individuals * self.best_proc))
                ),
                len(new_generation),
            )
            old_generation_n = self.num_individuals - new_generation_n

            population = (
                    population[:old_generation_n] + new_generation[:new_generation_n]
            )

            try:
                del new_generation
            except:
                pass

            gc.collect()

            population.sort(key=operator.attrgetter("fitness_value"), reverse=True)

            try:
                del population[self.num_individuals]
            except:
                pass

            # mutation params 12, 13
            # TODO: check this code
            for i in range(1, len(population)):
                #     if random.random() <= population[i].params[12]:
                #         for idx in range(3):
                #             if random.random() < population[i].params[13]:
                #                 if idx == 0:
                #                     population[i].params[12] = np.random.uniform(low=self.low_prob,
                #                                                                  high=self.high_prob, size=1)[0]
                #                 elif idx == 1:
                #                     population[i].params[13] = np.random.uniform(low=self.low_prob,
                #                                                                  high=self.high_prob, size=1)[0]
                #                 elif idx == 2:
                #                     population[i].params[13] = np.random.uniform(low=self.low_prob,
                #                                                                  high=self.high_prob, size=1)[0]

                if random.random() <= population[i].params[12]:
                    before_mutation.append(population[i])
                    id_mutation.append(i)
                    params = self.mutation(
                        copy.deepcopy(population[i].params),
                        elem_mutation_prob=copy.deepcopy(population[i].params[13]),
                        low_spb=self.low_spb,
                        high_spb=self.high_spb,
                        low_spm=self.low_spm,
                        high_spm=self.high_spm,
                        low_n=self.low_n,
                        high_n=self.high_n,
                        low_back=self.low_back,
                        high_back=self.high_back,
                        low_decor=self.low_decor,
                        high_decor=self.high_decor,
                    )

                    fix_value_13 = population[i].params[13]
                    for ix in [12, 13, 14]:
                        if random.random() < fix_value_13:
                            population[i].params[12] = np.random.uniform(
                                low=self.low_prob, high=self.high_prob, size=1
                            )[0]
                    params = self.check_params_bounds(params)
                    dto = IndividualDTO(
                        id=str(uuid.uuid4()),
                        data_path=self.data_path,
                        dataset=self.dataset,
                        params=[float(i) for i in params],
                        exp_id=self.exp_id,
                        alg_id=ALG_ID,
                        topic_count=self.topic_count,
                        tag=self.tag,
                        train_option=self.train_option,
                    )
                    population[i] = make_individual(dto=dto)
                self.evaluations_counter += 1

            # after the mutation we obtain a new population that needs to be evaluated
            for p in population:
                p.dto.iteration_id = ii

            fitness_calc_time_start = time.time()
            if not SPEEDUP or not self.surrogate:
                population = estimate_fitness(population)
                self.save_params(population)

            for ix, elem in enumerate(before_mutation):
                # TODO generation: int, original_params: list, mutated_params: list, original_fitness: float,
                #                       mutated_fitness: float
                self.metric_collector.save_mutation(
                    generation=ii,
                    original_params=elem.params,
                    mutated_params=population[id_mutation[ix]].params,
                    original_fitness=elem.fitness_value,
                    mutated_fitness=population[id_mutation[ix]].fitness_value,
                )

            logger.info(
                f"TIME OF THE FITNESS FUNCTION IN MUTATION: {time.time() - fitness_calc_time_start}"
            )

            if self.calc_scheme == "type1" and self.surrogate:
                if surrogate_iteration and self.surrogate:
                    self.surrogate_calculation(population)
                elif not surrogate_iteration and SPEEDUP and self.surrogate:
                    population = estimate_fitness(population)
                    self.save_params(population)
            elif self.calc_scheme == "type2" and self.surrogate:
                population = self._calculate_uncertain_res(population, iteration_num=ii)
                self.save_params(population)

            ###
            logger.info("MUTATION IS OVER")

            population.sort(key=operator.attrgetter("fitness_value"), reverse=True)

            if self.use_nelder_mead_in_mutation:
                collected_params = []
                for elem in population:
                    collected_params.append(elem.params)
                random_ids = random.sample(
                    [i for i in range(len(collected_params))], k=3
                )
                starting_points = [collected_params[i] for i in random_ids]

                nm_population = self.apply_nelder_mead(starting_points, num_gen=ii)
                for i, elem in enumerate(nm_population):
                    if population[i].fitness_value < elem.fitness_value:
                        print(
                            f"NM found better solution! {elem.fitness_value} vs {population[i].fitness_value}"
                        )
                        population[i] = elem

            if (
                    self.num_fitness_evaluations
                    and self.evaluations_counter >= self.num_fitness_evaluations
            ):
                self.metric_collector.save_fitness(
                    generation=ii,
                    params=[i.params for i in population],
                    fitness=[i.fitness_value for i in population],
                )
                bparams = "".join([str(i) for i in population[0].params])
                logger.info(
                    f"TERMINATION IS TRIGGERED: EVAL NUM (2)."
                    f"DATASET {self.dataset}."
                    f"TOPICS NUM {self.topic_count}."
                    f"RUN ID {run_id}."
                    f"THE BEST FITNESS {population[0].fitness_value}."
                    f"THE BEST PARAMS {bparams}."
                    f"ITERATION TIME {time.time() - iteration_start_time}."
                )
                # return population[0].fitness_value
                break

            current_fitness = population[0].fitness_value
            if (current_fitness > high_fitness) or (ii == 0):
                high_fitness = current_fitness

            if self.surrogate:
                if self.calc_scheme == "type1" and not surrogate_iteration:
                    self.surrogate.fit(
                        np.array(self.all_params), np.array(self.all_fitness)
                    )
                elif self.calc_scheme == "type2":
                    self.surrogate.fit(
                        np.array(self.all_params), np.array(self.all_fitness)
                    )

            if self.early_stopping_iterations:
                if population[0].fitness_value > best_val_so_far:
                    best_val_so_far = population[0].fitness_value
                    early_stopping_counter = 0
                else:
                    early_stopping_counter += 1
                    if early_stopping_counter == self.early_stopping_iterations:
                        bparams = "".join([str(i) for i in population[0].params])
                        self.metric_collector.save_fitness(
                            generation=ii,
                            params=[i.params for i in population],
                            fitness=[i.fitness_value for i in population],
                        )
                        logger.info(
                            f"TERMINATION IS TRIGGERED: EARLY STOPPING."
                            f"DATASET {self.dataset}."
                            f"TOPICS NUM {self.topic_count}."
                            f"RUN ID {run_id}."
                            f"THE BEST FITNESS {population[0].fitness_value}."
                            f"THE BEST PARAMS {bparams}."
                            f"ITERATION TIME {time.time() - iteration_start_time}."
                        )
                        break

            x.append(ii)
            y.append(population[0].fitness_value)
            self.metric_collector.save_fitness(
                generation=ii,
                params=[i.params for i in population],
                fitness=[i.fitness_value for i in population],
            )
            logger.info(
                f"Population len {len(population)}. "
                f"Best params so far: {population[0].params}, with fitness: {population[0].fitness_value}."
                f"ITERATION TIME: {time.time() - iteration_start_time}"
                f"DATASET {self.dataset}."
                f"TOPICS NUM {self.topic_count}."
                f"RUN ID {run_id}."
            )
            best_solution = population[0]
            log_best_solution(best_solution, alg_args=" ".join(sys.argv), is_tmp=True)

        self.metric_collector.save_and_visualise_trace()

        logger.info(f"Y: {y}")
        best_individual = population[0]
        ind = log_best_solution(best_individual, alg_args=" ".join(sys.argv))
        logger.info(
            f"Logged the best solution. Obtained fitness is {ind.fitness_value}"
        )

        return ind.fitness_value


# multistage bag of regularizers approach
# TODO: add penalty for the solution length


class GAmultistage(GA):
    def __init__(self, dataset, num_individuals, max_stages=5):  # max_stage_len
        self.max_stages = max_stages  # amount of unique regularizers
        self.dataset = dataset
        self.bag_of_regularizers = [
            "decor_S",
            "decor_B",
            "S_phi_B",
            "S_phi_S",
            "S_theta_B",
            "S_theta_S",
        ]  # add separate smooth and sparsity
        self.num_individuals = num_individuals
        self.initial_element_stage_probability = 0.5
        self.positioning_matrix = np.full(
            (len(self.bag_of_regularizers), self.max_stages - 1), 0.5
        )
        self.set_regularizer_limits()

    def set_regularizer_limits(
            self,
            low_decor=0,
            high_decor=1e5,
            low_n=1,
            high_n=30,  # minimal value changed to 1
            low_back=0,
            high_back=5,
            low_spb=0,
            high_spb=1e2,
            low_spm=-1e-3,
            high_spm=1e2,
            low_sp_phi=-1e3,
            high_sp_phi=1e3,
    ):
        self.high_decor = high_decor
        self.low_decor = low_decor
        self.low_n = low_n
        self.high_n = high_n
        self.low_back = low_back
        self.high_back = high_back
        self.high_spb = high_spb
        self.low_spb = low_spb
        self.low_spm = low_spm
        self.high_spm = high_spm
        self.low_sp_phi = low_sp_phi
        self.high_sp_phi = high_sp_phi

    # TODO: check if float is needed
    def _init_param(self, param_type):
        if param_type == "decor_S" or param_type == "decor_B":
            return float(
                np.random.uniform(low=self.low_decor, high=self.high_decor, size=1)[0]
            )
        elif param_type == "S_phi_B" or "S_theta_B":
            return float(
                np.random.uniform(low=self.low_spb, high=self.high_spb, size=1)[0]
            )
        elif param_type == "S_phi_S" or "S_theta_S":
            return float(
                np.random.uniform(low=self.low_sp_phi, high=self.high_sp_phi, size=1)[0]
            )
        elif param_type == "n":
            return float(np.random.randint(low=self.low_n, high=self.high_n, size=1)[0])
        elif param_type == "B":
            return float(
                np.random.randint(low=self.low_back, high=self.high_back, size=1)[0]
            )

    def _create_stage(self, stage_num):
        stage_regularizers = []
        for ix, elem in enumerate(self.bag_of_regularizers):
            elem_sample_prob = self.positioning_matrix[ix][stage_num - 1]
            if random.random() < elem_sample_prob:
                stage_regularizers.append(elem)
        stage_regularizers.append("n")
        return stage_regularizers

    def init_individ(self):
        number_of_stages = np.random.randint(low=1, high=self.max_stages, size=1)[0]
        dict_of_stages = [{} for _ in range(number_of_stages)]
        for i in range(number_of_stages):
            regularizers = self._create_stage(i)
            for reg_name in regularizers:
                value = self._init_param(reg_name)
                dict_of_stages[i][reg_name] = value
        dict_of_stages = [{"B": self._init_param("B")}] + dict_of_stages
        return dict_of_stages

    def init_population(self):
        # if random.random() < self.initial_element_stage_probability:
        #
        # for i in range(self.max_stages):
        #     print()

        raise NotImplementedError



from scipy.optimize import minimize
from typing import Union
from autotm.fitness.tm import FitnessCalculatorWrapper
import numpy as np


class NelderMeadOptimization:
    def __init__(
        self,
        dataset,
        data_path,
        exp_id,
        topic_count,
        train_option,
        low_decor=0,
        high_decor=1e5,
        low_n=0,
        high_n=30,
        low_back=0,
        high_back=5,
        low_spb=0,
        high_spb=1e2,
        low_spm=-1e-3,
        high_spm=1e2,
        low_sp_phi=-1e3,
        high_sp_phi=1e3,
        low_prob=0,
        high_prob=1,
    ):
        self.dataset = (dataset,)
        self.data_path = data_path
        self.exp_id = exp_id
        self.topic_count = topic_count
        self.train_option = train_option
        self.high_decor = high_decor
        self.low_decor = low_decor
        self.low_n = low_n
        self.high_n = high_n
        self.low_back = low_back
        self.high_back = high_back
        self.high_spb = high_spb
        self.low_spb = low_spb
        self.low_spm = low_spm
        self.high_spm = high_spm
        self.low_sp_phi = low_sp_phi
        self.high_sp_phi = high_sp_phi
        self.low_prob = low_prob
        self.high_prob = high_prob

    def initialize_params(self):
        val_decor = np.random.uniform(low=self.low_decor, high=self.high_decor, size=1)[
            0
        ]
        var_n = np.random.randint(low=self.low_n, high=self.high_n, size=4)
        var_back = np.random.randint(low=self.low_back, high=self.high_back, size=1)[0]
        var_sm = np.random.uniform(low=self.low_spb, high=self.high_spb, size=2)
        var_sp = np.random.uniform(low=self.low_sp_phi, high=self.high_sp_phi, size=4)
        val_decor_2 = np.random.uniform(
            low=self.low_decor, high=self.high_decor, size=1
        )[0]
        params = [
            val_decor,
            var_n[0],
            var_sm[0],
            var_sm[1],
            var_n[1],
            var_sp[0],
            var_sp[1],
            var_n[2],
            var_sp[2],
            var_sp[3],
            var_n[3],
            var_back,
            val_decor_2,
        ]
        params = [float(i) for i in params]
        return params

    def run_algorithm(self, num_iterations: int = 400, ini_point: list = None):
        fitness_calculator = FitnessCalculatorWrapper(
            self.dataset, self.data_path, self.topic_count, self.train_option
        )

        if ini_point is None:
            initial_point = self.initialize_params()
        else:
            assert len(ini_point) == 13
            print(ini_point)  # TODO: remove this
            ini_point = [float(i) for i in ini_point]
            initial_point = ini_point

        res = minimize(
            fitness_calculator.run,
            initial_point,
            bounds=[
                (self.low_decor, self.high_decor),
                (self.low_n, self.high_n),
                (self.low_spb, self.high_spb),
                (self.low_spb, self.high_spb),
                (self.low_n, self.high_n),
                (self.low_sp_phi, self.high_sp_phi),
                (self.low_sp_phi, self.high_sp_phi),
                (self.low_n, self.high_n),
                (self.low_sp_phi, self.high_sp_phi),
                (self.low_sp_phi, self.high_sp_phi),
                (self.low_n, self.high_n),
                (self.low_back, self.high_back),
                (self.low_decor, self.high_decor),
            ],
            method="Nelder-Mead",
            options={"return_all": True, "maxiter": num_iterations},
        )

        return res



import logging
from typing import List, Optional

from autotm.params_logging_utils import log_params_and_artifacts, log_stats, model_files
from autotm.fitness.tm import fit_tm_of_individual
from autotm.schemas import IndividualDTO, fitness_to_json, fitness_from_json
from autotm.algorithms_for_tuning.individuals import make_individual

logger = logging.getLogger("root")


def do_fitness_calculating(
        individual: str,
        log_artifact_and_parameters: bool = False,
        log_run_stats: bool = False,
        alg_args: Optional[str] = None,
        is_tmp: bool = False,
) -> str:
    individual: IndividualDTO = IndividualDTO.parse_raw(individual)

    with fit_tm_of_individual(
            dataset=individual.dataset,
            data_path=individual.data_path,
            params=individual.params,
            fitness_name=individual.fitness_name,
            topic_count=individual.topic_count,
            force_dataset_settings_checkout=individual.force_dataset_settings_checkout,
            train_option=individual.train_option,
    ) as (time_metrics, metrics, tm):
        individual.fitness_value = metrics

        with model_files(tm) as tm_files:
            if log_artifact_and_parameters:
                log_params_and_artifacts(
                    tm, tm_files, individual, time_metrics, alg_args, is_tmp=is_tmp
                )

            if log_run_stats:
                log_stats(tm, tm_files, individual, time_metrics, alg_args)

    return individual.json()


def calculate_fitness(
        individual: str,
        log_artifact_and_parameters: bool = False,
        log_run_stats: bool = False,
        alg_args: Optional[str] = None,
        is_tmp: bool = False
) -> str:
    try:
        return do_fitness_calculating(
            individual, log_artifact_and_parameters, log_run_stats, is_tmp
        )
    except Exception as e:
        print(str(e))
        raise Exception(f"Some exception")


def estimate_fitness(population: List[IndividualDTO]) -> List[IndividualDTO]:
    logger.info("Calculating fitness...")
    population_with_fitness = []
    for individual in population:
        json_individ = fitness_to_json(individual.dto)
        individ_with_fitness = calculate_fitness(json_individ)
        population_with_fitness.append(
            make_individual(fitness_from_json(individ_with_fitness))
        )
    logger.info("The fitness results have been obtained")
    return population_with_fitness


def log_best_solution(
        individual: IndividualDTO,
        wait_for_result_timeout: Optional[float] = None,
        alg_args: Optional[str] = None,
        is_tmp: bool = False,
):
    logger.info(f"Sending a best individual to be logged")
    res = make_individual(
        fitness_from_json(
            calculate_fitness(
                fitness_to_json(individual.dto), log_artifact_and_parameters=True, is_tmp=is_tmp
            )
        )
    )

    # TODO: write logging
    return res



import random
import numpy as np


def mutation_one_param(
    individ,
    low_spb,
    high_spb,
    low_spm,
    high_spm,
    low_n,
    high_n,
    low_back,
    high_back,
    low_decor,
    high_decor,
    elem_mutation_prob=0.1,
):
    for i in range(len(individ)):
        if random.random() <= elem_mutation_prob:
            if i in [2, 3]:
                individ[i] = np.random.uniform(low=low_spb, high=high_spb, size=1)[0]
            for i in [5, 6, 8, 9]:
                individ[i] = np.random.uniform(low=low_spm, high=high_spm, size=1)[0]
            for i in [1, 4, 7, 10]:
                individ[i] = float(np.random.randint(low=low_n, high=high_n, size=1)[0])
            for i in [11]:
                individ[i] = float(
                    np.random.randint(low=low_back, high=high_back, size=1)[0]
                )
            for i in [0, 15]:
                individ[i] = np.random.uniform(low=low_decor, high=high_decor, size=1)[
                    0
                ]
    return individ


def positioning_mutation(individ, elem_mutation_prob=0.1, **kwargs):
    for i in range(len(individ)):
        set_1 = set([i])
        if random.random() <= elem_mutation_prob:
            if i in [0, 15]:
                set_2 = set([0, 15])
                ix = np.random.choice(list(set_2.difference(set_1)))
                tmp = individ[ix]
                individ[ix] = individ[i]
                individ[i] = tmp
            elif i in [1, 4, 7, 10, 11]:
                set_2 = set([1, 4, 7, 10, 11])
                ix = np.random.choice(list(set_2.difference(set_1)))
                tmp = individ[ix]
                individ[ix] = individ[i]
                individ[i] = tmp
            elif i in [2, 3]:
                set_2 = set([2, 3])
                ix = np.random.choice(list(set_2.difference(set_1)))
                tmp = individ[ix]
                individ[ix] = individ[i]
                individ[i] = tmp
            elif i in [5, 6, 8, 9]:
                set_2 = set([5, 6, 8, 9])
                ix = np.random.choice(list(set_2.difference(set_1)))
                tmp = individ[ix]
                individ[ix] = individ[i]
                individ[i] = tmp
        return individ


def mutation_combined(individ, elem_mutation_prob=0.1, **kwargs):
    if random.random() <= individ[14]:  # TODO: check 14th position
        return mutation_one_param(individ, elem_mutation_prob)
    else:
        return positioning_mutation(individ, elem_mutation_prob)
    pass


def mutation_psm_new(individ, elem_mutation_prob, **kwargs):
    for i in range(len(individ)):
        if random.random() < elem_mutation_prob:
            if i == 0:
                individ[i] = individ[15]
            elif i == 15:
                individ[i] = individ[0]
            elif i in [1, 4, 7, 10]:
                j = np.random.choice([1, 4, 7, 10])
                tmp = individ[i]
                individ[i] = individ[j]
                individ[j] = tmp
            elif i in [2, 3]:
                j = np.random.choice([2, 3])
                tmp = individ[i]
                individ[i] = individ[j]
                individ[j] = tmp
            # elif i in [12, 13, 14]:
            #     j = np.random.choice([12, 13, 14])
            #     tmp = individ[i]
            #     individ[i] = individ[j]
            #     individ[j] = tmp
            elif i in [5, 6, 8, 9]:
                j = np.random.choice([5, 6, 8, 9])
                tmp = individ[i]
                individ[i] = individ[j]
                individ[j] = tmp
    return individ


def mutation_psm(individ, elem_mutation_prob=0.1, **kwargs):
    for i in range(len(individ)):
        if random.random() < elem_mutation_prob:
            if i == 0:
                individ[i] = np.random.uniform(low=1, high=100, size=1)[0]
            elif i in [1, 4, 7]:
                j = np.random.choice([1, 4, 7])
                tmp = individ[i]
                individ[i] = individ[j]
                individ[j] = tmp
            elif i in [2, 5]:
                j = np.random.choice([2, 5])
                tmp = individ[i]
                individ[i] = individ[j]
                individ[j] = tmp
            elif i in [3, 6]:
                j = np.random.choice([2, 5])
                tmp = individ[i]
                individ[i] = individ[j]
                individ[j] = tmp
    return individ


def mutation(mutation_type="mutation_one_param"):
    if mutation_type == "mutation_one_param":
        return mutation_one_param
    if mutation_type == "combined":
        return mutation_combined
    if mutation_type == "psm":
        return mutation_psm
    if mutation_type == "positioning_mutation":
        return positioning_mutation



#!/usr/bin/python3
import datetime
import logging
import os
import shutil
import sys
from glob import glob

logging.basicConfig(
    level=logging.DEBUG, format="%(asctime)s - %(name)s - %(levelname)s - %(message)s"
)
logger = logging.getLogger()


# example: irace-log-2021-07-19T21:48:00-2b5faba3-99a8-47ba-a5cd-2c99e6877160.Rdata
def fpath_to_datetime(fpath: str) -> datetime:
    fname = os.path.basename(fpath)
    dt_str = fname[len("irace-log-") : len("irace-log-") + len("YYYY-mm-ddTHH:MM:ss")]
    dt = datetime.datetime.strptime(dt_str, "%Y-%m-%dT%H:%M:%S")
    return dt


if __name__ == "__main__":
    save_dir = sys.argv[1]
    logger.info(f"Looking for data files in {save_dir}")
    data_files = [
        (fpath, fpath_to_datetime(fpath)) for fpath in glob(f"{save_dir}/*.Rdata")
    ]

    if len(data_files) == 0:
        logger.info(
            f"No data files have been found in {save_dir}. It is a normal situation. Interrupting execution."
        )
        sys.exit(0)

    last_data_filepath, _ = max(data_files, key=lambda x: x[1])
    recovery_filepath = os.path.join(save_dir, "recovery_rdata.checkpoint")

    logger.info(
        f"Copying last generated data file ({last_data_filepath}) as recovery file ({recovery_filepath})"
    )
    shutil.copy(last_data_filepath, recovery_filepath)
    logger.info(f"Copying done")



import copy
import json
from typing import List, Optional
from pydantic import BaseModel

from autotm.utils import MetricsScores

PARAM_NAMES = [
    "val_decor",
    "var_n_0",
    "var_sm_0",
    "var_sm_1",
    "var_n_1",
    "var_sp_0",
    "var_sp_1",
    "var_n_2",
    "var_sp_2",
    "var_sp_3",
    "var_n_3",
    "var_n_4",
    "ext_mutation_prob",
    "ext_elem_mutation_prob",
    "ext_mutation_selector",
    "val_decor_2",
]


class IndividualDTO(BaseModel):
    id: str
    data_path: str
    params: List[object]
    fitness_name: str = "default"
    dataset: str = "default"
    force_dataset_settings_checkout: bool = False
    fitness_value: MetricsScores = None
    exp_id: Optional[int] = None
    alg_id: Optional[str] = None
    tag: Optional[str] = None
    iteration_id: int = 0
    topic_count: Optional[int] = None
    train_option: str = "offline"

    class Config:
        arbitrary_types_allowed = True

    def make_params_dict(self):
        if len(self.params) > len(PARAM_NAMES):
            len_diff = len(self.params) - len(PARAM_NAMES)
            param_names = copy.deepcopy(PARAM_NAMES) + [
                f"unknown_param_#{i}" for i in range(len_diff)
            ]
        else:
            param_names = PARAM_NAMES

        return {name: p_val for name, p_val in zip(param_names, self.params)}


def fitness_to_json(obj: IndividualDTO):
    return json.dumps(
        {
            "id": obj.id,
            "data_path": obj.data_path,
            "params": obj.params,
            "exp_id": obj.exp_id,
            "alg_id": obj.alg_id,
            "iteration_id": obj.iteration_id,
            "tag": obj.tag,
            "fitness_value": obj.fitness_value,
            "fitness_name": obj.fitness_name,
            "dataset": obj.dataset,
            "topic_count": obj.topic_count,
            "train_option": obj.train_option,
            "force_dataset_settings_checkout": obj.force_dataset_settings_checkout,
        }
    )


def fitness_from_json(obj):
    obj = json.loads(obj)
    return IndividualDTO(
        id=obj["id"],
        data_path=obj["data_path"],
        params=obj["params"],
        exp_id=int(obj["exp_id"]) if obj["exp_id"] else None,
        alg_id=obj["alg_id"] if obj["alg_id"] else None,
        iteration_id=obj["iteration_id"],
        tag=obj["tag"],
        fitness_value=obj["fitness_value"],
        fitness_name=obj["fitness_name"],
        dataset=obj["dataset"],
        topic_count=obj["topic_count"],
        train_option=obj["train_option"],
        force_dataset_settings_checkout=obj["force_dataset_settings_checkout"],
    )



from abc import ABC, abstractmethod
import os
import pickle
from typing import List
import numpy as np
import pandas as pd

from autotm.utils import AVG_COHERENCE_SCORE
from autotm.schemas import IndividualDTO

SPARSITY_PHI = "sparsity_phi"
SPARSITY_THETA = "sparsity_theta"
SWITCHP_SCORE = "switchP"
DF_NAMES = {"20ng": 0, "lentaru": 1, "amazon_food": 2}

METRICS_COLS = [
    "avg_coherence_score",
    "perplexityScore",
    "backgroundTokensRatioScore",
    "avg_switchp",
    "coherence_10",
    "coherence_15",
    "coherence_20",
    "coherence_25",
    "coherence_30",
    "coherence_35",
    "coherence_40",
    "coherence_45",
    "coherence_50",
    "coherence_55",
    "contrast",
    "purity",
    "kernelSize",
    "sparsity_phi",
    "sparsity_theta",
    "topic_significance_uni",
    "topic_significance_vacuous",
    "topic_significance_back",
    "npmi_15",
    "npmi_25",
    "npmi_50",
]

PATH_TO_LEARNED_SCORING = "./scoring_func"


class Individual(ABC):
    @property
    @abstractmethod
    def dto(self) -> IndividualDTO:
        ...

    @property
    @abstractmethod
    def fitness_value(self) -> float:
        ...

    @property
    @abstractmethod
    def params(self) -> List:
        ...


class BaseIndividual(Individual, ABC):
    def __init__(self, dto: IndividualDTO):
        self._dto = dto

    @property
    def dto(self) -> IndividualDTO:
        return self._dto

    @property
    def params(self) -> List:
        return self.dto.params


class RegularFitnessIndividual(BaseIndividual):
    @property
    def fitness_value(self) -> float:
        return self.dto.fitness_value[AVG_COHERENCE_SCORE]


class LearnedModel:
    def __init__(self, save_path, dataset_name):
        dataset_id = DF_NAMES[dataset_name]
        general_save_path = os.path.join(save_path, "general")
        native_save_path = os.path.join(save_path, "native")
        with open(
            os.path.join(general_save_path, f"general_automl_{dataset_id}.pickle"), "rb"
        ) as f:
            self.general_model = pickle.load(f)
        self.native_model = []
        for i in range(5):
            with open(
                os.path.join(
                    native_save_path, f"native_automl_{dataset_id}_fold_{i}.pickle"
                ),
                "rb",
            ) as f:
                self.native_model.append(pickle.load(f))

    def general_predict(self, df: pd.DataFrame):
        y = self.general_model.predict(df[METRICS_COLS])
        return y

    def native_predict(self, df: pd.DataFrame):
        y = []
        for k, nm in enumerate(self.native_model):
            y.append(nm.predict(df[METRICS_COLS]))
        y = np.array(y)
        return np.mean(y, axis=0)


class LearnedScorerFitnessIndividual(BaseIndividual):
    @property
    def fitness_value(self) -> float:
        dataset_name = self.dto.dataset  # TODO: check namings
        m = LearnedModel(save_path=PATH_TO_LEARNED_SCORING, dataset_name=dataset_name)
        # TODO: predict from metrics df


class SparsityScalerBasedFitnessIndividual(BaseIndividual):
    @property
    def fitness_value(self) -> float:
        # it is a handling of the situation when a fitness-worker wasn't able to correctly calculate this indvidual
        # due to some error in the proceess
        # and thus the fitness value doesn't have any metrics except dummy AVG_COHERENCE_SCORE equal to zero
        if self.dto.fitness_value[AVG_COHERENCE_SCORE] < 0.00000001:
            return 0.0

        alpha = 0.7
        if 0.2 <= self.dto.fitness_value[SPARSITY_THETA] <= 0.8:
            alpha = 1
        # if SWITCHP_SCORE in self.dto.fitness_value:
        #     return alpha * (self.dto.fitness_value[AVG_COHERENCE_SCORE] + self.dto.fitness_value[SWITCHP_SCORE])
        # else:
        #     return alpha * self.dto.fitness_value[AVG_COHERENCE_SCORE]
        return alpha * self.dto.fitness_value[AVG_COHERENCE_SCORE]


def make_individual(dto: IndividualDTO) -> Individual:
    # TODO: choose fitness by ENV var
    return RegularFitnessIndividual(dto=dto)
    # return SparsityScalerBasedFitnessIndividual(dto=dto)



import os
from artm.batches_utils import BatchVectorizer, Batch
import glob
import random


class SampleBatchVectorizer(BatchVectorizer):
    def __init__(self, sample_size=10, **kwargs):
        self.sample_size = sample_size
        super().__init__(**kwargs)

    def _parse_batches(self, data_weight, batches):
        if self._process_in_memory:
            self._model.master.import_batches(batches)
            self._batches_list = [batch.id for batch in batches]
            return

        data_paths, data_weights, target_folders = self._populate_data(
            data_weight, True
        )
        for data_p, data_w, target_f in zip(data_paths, data_weights, target_folders):
            if batches is None:
                batch_filenames = glob.glob(os.path.join(data_p, "*.batch"))
                if len(batch_filenames) < self.sample_size:
                    self.sample_size = len(batch_filenames)
                batch_filenames = random.sample(batch_filenames, self.sample_size)
                self._batches_list += [Batch(filename) for filename in batch_filenames]

                if len(self._batches_list) < 1:
                    raise RuntimeError("No batches were found")

                self._weights += [data_w for i in range(len(batch_filenames))]
            else:
                self._batches_list += [
                    Batch(os.path.join(data_p, batch)) for batch in batches
                ]
                self._weights += [data_w for i in range(len(batches))]



import os
import yaml
import json

import pandas as pd
import artm
from autotm.preprocessing.dictionaries_preparation import prepare_batch_vectorizer

PATH_TO_RUN_NAME = "tags/mlflow.runName"
PATH_TO_EXPERIMENT_ID = ""

PROCESSED_TEXT_COL = "processed_text"
TOP_TOPICS_COL = "SER_top_topics"


def get_experiment_path(
        exp_id: int, run_name: str, mlflow_path: str = "./mlruns/"
) -> str:
    """

    :param mlflow_path: path where mlflow stores the results default is
    :param exp_id: id of the experiment, which artifacts is needed
    :param run_name: name of the mlflow run
    :return: path to the experiment artifacts
    """
    for folder in os.listdir(mlflow_path):
        if folder != ".trash":
            subfolders = os.listdir(os.path.join(mlflow_path, folder))
            if len(subfolders) > 1:
                with open(os.path.join(mlflow_path, folder, "meta.yaml"), "r") as f:
                    data = yaml.load(f, Loader=yaml.Loader)
                if data["name"] == f"experiment_{exp_id}":
                    for subfolder in subfolders:
                        if subfolder != "meta.yaml":
                            with open(
                                    os.path.join(
                                        mlflow_path, folder, subfolder, PATH_TO_RUN_NAME
                                    )
                            ) as f:
                                f_run_name = f.read()
                                if f_run_name.strip() == run_name:
                                    return os.path.join(
                                        mlflow_path, folder, subfolder, "artifacts"
                                    )
    return None


def get_artifacts(artifacts_path: str):
    phi_folders = os.listdir(os.path.join(artifacts_path, "phi.csv"))
    theta_folders = os.listdir(os.path.join(artifacts_path, "theta.csv"))
    phi_matrix = pd.read_csv(os.path.join(artifacts_path, "phi.csv", phi_folders[0]))
    theta_matrix = pd.read_csv(
        os.path.join(artifacts_path, "theta.csv", theta_folders[0])
    )
    with open(os.path.join(artifacts_path, "topics.json")) as f:
        topics = json.load(f)
    return phi_matrix, theta_matrix, topics


# def _get_phi_dict_format(df):
#


def _transform_matrix(df: pd.DataFrame) -> pd.DataFrame:
    if "Unnamed: 0" in list(df):
        df = df.set_index("Unnamed: 0")
        df = df.T
    return df


def get_most_probable_topics_from_theta(
        df: pd.DataFrame, theta_df: pd.DataFrame, top_n: int = 2
) -> pd.DataFrame:
    """

    :param df: processed dataset which is produced by prepare_all function
    :param theta_df: theta matrix of trained model
    :param top_n: amount of top topics to consider
    :return: pd.Dataframe with 'top_topics' column
    """
    theta_df = _transform_matrix(theta_df)
    print(theta_df)
    assert (
            df.shape[0] == theta_df.shape[0]
    ), "Shapes of f and theta matrix are different"
    print(theta_df.apply(lambda x: ", ".join(x.nlargest(top_n).index.tolist()), axis=1))
    df[TOP_TOPICS_COL] = theta_df.apply(
        lambda x: ", ".join(x.nlargest(top_n).index.tolist()), axis=1
    ).tolist()
    return df


def get_top_words_from_topic_in_text(df: pd.DataFrame, topics: dict, top_w: int = 2):
    assert TOP_TOPICS_COL in list(
        df
    ), "use function get_most_probable_topics_from_theta to get top topics for texts"
    all_dicts = []
    for id, row in df.iterrows():
        topic_tokens = {}
        top_topics_names = [i.strip() for i in row[TOP_TOPICS_COL].split(",")]
        for topic in top_topics_names:
            words = set(row[PROCESSED_TEXT_COL].split()).intersection(
                set(topics[topic])
            )
            if len(words) > 1:
                topic_tokens[topic] = [word for word in topics[topic] if word in words][
                                      :top_w
                                      ]
        all_dicts.append(topic_tokens)
    df["SER_words_of_topic"] = all_dicts
    return df


def get_most_probable_words_from_phi(df, phi_df):
    raise NotImplementedError


class TopicsExtractor:
    def __init__(self, model_path):
        self.__tm_model_path = model_path
        self.model = self.__load_model()
        self.__tmp_batches_path = "./tmp/tmp_batches/"
        self.__tmp_dict_path = "./tmp/tmp_dict/"
        self.topics_dict = self.model.score_tracker["TopTokensScore"].last_tokens

    def info(self):
        pass

    def return_topics(self):
        pass

    def __load_model(self):
        return artm.load_artm_model(self.__tm_model_path)

    def get_prob_mixture(
            self,
            data_path,
            TMP_BATCHES_PATH="./tmp/tmp_batches/",
            TMP_DICT_PATH="./tmp/tmp_dict/",
            OUTPUT_DIR="./out",
            text_column_name="processed_text",
            input_format="csv",
            top_n=2,
    ):
        try:
            os.makedirs(TMP_BATCHES_PATH)
            os.makedirs(TMP_DICT_PATH)
        except:
            print("folders already exist")

        for tmp_batch in os.listdir(TMP_BATCHES_PATH):
            os.remove(os.path.join(TMP_BATCHES_PATH, tmp_batch))

        if input_format == "csv":
            posts = pd.read_csv(data_path)
        else:
            posts = pd.read_parquet(data_path)

        posts = posts[~posts[text_column_name].isna()]
        posts = posts.reset_index(drop=True)

        texts = posts[text_column_name].tolist()

        print("LEN {}".format(len(texts)))

        batch_vectorizer_test = prepare_batch_vectorizer(
            batches_dir=TMP_DICT_PATH,
            vw_path="./tmp/tmp_wv.txt",
            data_path=data_path,
            column_name=text_column_name,
        )

        theta_test = self.model.transform(batch_vectorizer=batch_vectorizer_test)
        theta_test_trans = theta_test.T
        main_topics = [i for i in list(theta_test_trans) if i.startswith("main")]
        theta_test_trans["top_topics"] = (
            theta_test_trans[main_topics]
            .apply(lambda x: ", ".join(x.nlargest(top_n).index.tolist()), axis=1)
            .tolist()
        )
        theta_test_trans = theta_test_trans.join(posts[[text_column_name]])
        theta_test_trans.to_csv(
            os.path.join(OUTPUT_DIR, "data_with_theta.csv"), index=None
        )

        print("All is saved to {} !".format(OUTPUT_DIR))



# strategies of automatic EA configuration

# CMA-ES (Covariance Matrix Adaptation Evolution Strategy)

# code source: https://github.com/DEAP/deap/blob/master/deap/cma.py
import numpy
from math import sqrt, log, exp


class Strategy(object):
    def __init__(self, centroid, sigma, **kwargs):
        self.params = kwargs

        # Create a centroid as a numpy array
        self.centroid = numpy.array(centroid)

        self.dim = len(self.centroid)
        self.sigma = sigma
        self.pc = numpy.zeros(self.dim)
        self.ps = numpy.zeros(self.dim)
        self.chiN = sqrt(self.dim) * (
            1 - 1.0 / (4.0 * self.dim) + 1.0 / (21.0 * self.dim**2)
        )

        self.C = self.params.get("cmatrix", numpy.identity(self.dim))
        self.diagD, self.B = numpy.linalg.eigh(self.C)

        indx = numpy.argsort(self.diagD)
        self.diagD = self.diagD[indx] ** 0.5
        self.B = self.B[:, indx]
        self.BD = self.B * self.diagD

        self.cond = self.diagD[indx[-1]] / self.diagD[indx[0]]

        self.lambda_ = self.params.get("lambda_", int(4 + 3 * log(self.dim)))
        self.update_count = 0
        self.computeParams(self.params)

    def generate(self, ind_init):
        r"""Generate a population of :math:`\lambda` individuals of type
        *ind_init* from the current strategy.
        :param ind_init: A function object that is able to initialize an
                         individual from a list.
        :returns: A list of individuals.
        """
        arz = numpy.random.standard_normal((self.lambda_, self.dim))
        arz = self.centroid + self.sigma * numpy.dot(arz, self.BD.T)
        return [ind_init(a) for a in arz]

    def update(self, population):
        """Update the current covariance matrix strategy from the
        *population*.
        :param population: A list of individuals from which to update the
                           parameters.
        """
        population.sort(key=lambda ind: ind.fitness, reverse=True)

        old_centroid = self.centroid
        self.centroid = numpy.dot(self.weights, population[0 : self.mu])

        c_diff = self.centroid - old_centroid

        # Cumulation : update evolution path
        self.ps = (1 - self.cs) * self.ps + sqrt(
            self.cs * (2 - self.cs) * self.mueff
        ) / self.sigma * numpy.dot(
            self.B, (1.0 / self.diagD) * numpy.dot(self.B.T, c_diff)
        )

        hsig = float(
            (
                numpy.linalg.norm(self.ps)
                / sqrt(1.0 - (1.0 - self.cs) ** (2.0 * (self.update_count + 1.0)))
                / self.chiN
                < (1.4 + 2.0 / (self.dim + 1.0))
            )
        )

        self.update_count += 1

        self.pc = (1 - self.cc) * self.pc + hsig * sqrt(
            self.cc * (2 - self.cc) * self.mueff
        ) / self.sigma * c_diff

        # Update covariance matrix
        artmp = population[0 : self.mu] - old_centroid
        self.C = (
            (
                1
                - self.ccov1
                - self.ccovmu
                + (1 - hsig) * self.ccov1 * self.cc * (2 - self.cc)
            )
            * self.C
            + self.ccov1 * numpy.outer(self.pc, self.pc)
            + self.ccovmu * numpy.dot((self.weights * artmp.T), artmp) / self.sigma**2
        )

        self.sigma *= numpy.exp(
            (numpy.linalg.norm(self.ps) / self.chiN - 1.0) * self.cs / self.damps
        )

        self.diagD, self.B = numpy.linalg.eigh(self.C)
        indx = numpy.argsort(self.diagD)

        self.cond = self.diagD[indx[-1]] / self.diagD[indx[0]]

        self.diagD = self.diagD[indx] ** 0.5
        self.B = self.B[:, indx]
        self.BD = self.B * self.diagD

    def computeParams(self, params):
        r"""Computes the parameters depending on :math:`\lambda`. It needs to
        be called again if :math:`\lambda` changes during evolution.
        :param params: A dictionary of the manually set parameters.
        """
        self.mu = params.get("mu", int(self.lambda_ / 2))
        rweights = params.get("weights", "superlinear")
        if rweights == "superlinear":
            self.weights = log(self.mu + 0.5) - numpy.log(numpy.arange(1, self.mu + 1))
        elif rweights == "linear":
            self.weights = self.mu + 0.5 - numpy.arange(1, self.mu + 1)
        elif rweights == "equal":
            self.weights = numpy.ones(self.mu)
        else:
            raise RuntimeError("Unknown weights : %s" % rweights)

        self.weights /= sum(self.weights)
        self.mueff = 1.0 / sum(self.weights**2)

        self.cc = params.get("ccum", 4.0 / (self.dim + 4.0))
        self.cs = params.get("cs", (self.mueff + 2.0) / (self.dim + self.mueff + 3.0))
        self.ccov1 = params.get("ccov1", 2.0 / ((self.dim + 1.3) ** 2 + self.mueff))
        self.ccovmu = params.get(
            "ccovmu",
            2.0
            * (self.mueff - 2.0 + 1.0 / self.mueff)
            / ((self.dim + 2.0) ** 2 + self.mueff),
        )
        self.ccovmu = min(1 - self.ccov1, self.ccovmu)
        self.damps = (
            1.0
            + 2.0 * max(0, sqrt((self.mueff - 1.0) / (self.dim + 1.0)) - 1.0)
            + self.cs
        )
        self.damps = params.get("damps", self.damps)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['autotm',
 'autotm.algorithms_for_tuning',
 'autotm.algorithms_for_tuning.bayesian_optimization',
 'autotm.algorithms_for_tuning.genetic_algorithm',
 'autotm.algorithms_for_tuning.nelder_mead_optimization',
 'autotm.fitness',
 'autotm.preprocessing',
 'autotm.visualization']

package_data = \
{'': ['*'], 'autotm': ['data_generator/*']}

install_requires = \
['PyYAML',
 'bigartm==0.9.2',
 'billiard',
 'click',
 'dataclasses-json',
 'dill',
 'gensim',
 'hyperopt',
 'mlflow',
 'nltk',
 'numpy',
 'pandas',
 'plotly',
 'pydantic',
 'pymongo',
 'pymystem3',
 'pytest',
 'scikit-learn',
 'scipy',
 'spacy',
 'spacy-langdetect',
 'tqdm']

setup_kwargs = {
    'name': 'autotm',
    'version': '0.1.0',
    'description': 'Automatic hyperparameters tuning for topic models (ARTM approach) using evolutionary algorithms',
    'long_description': '# AutoTM\n\n[![Project Status: WIP – Initial development is in progress, but there has not yet been a stable, usable release suitable for the public.](https://www.repostatus.org/badges/latest/wip.svg)](https://www.repostatus.org/#wip)\n![build](https://github.com/ngc436/AutoTM/actions/workflows/build.yaml/badge.svg)\n[![GitHub Repo stars](https://img.shields.io/github/stars/ngc436/AutoTM?style=social)](https://github.com/ngc436/AutoTM/stargazers)\n\nAutomatic parameters selection for topic models (ARTM approach) using evolutionary algorithms. \nAutoTM provides necessary tools to preprocess english and russian text datasets and tune topic models.\n\n## What is AutoTM?\nTopic modeling is one of the basic methods for EDA of unlabelled text data. While ARTM (additive regularization \nfor topic models) approach provides the significant flexibility and quality comparative or better that neural \napproaches it is hard to tune such models due to amount of hyperparameters and their combinations.\n\nTo overcome the tuning problems AutoTM presents an easy way to represent a learning strategy to train specific models for input corporas.\n\n<img src="docs/img/strategy.png" alt="Learning strategy representation" height=""/>\n\nOptimization procedure is done by genetic algorithm which operators are specifically tuned for \nthe task. To speed up the procedure we also implemented surrogate modeling that, for some iterations, \napproximate fitness function to reduce computation costs on training topic models.\n\n<img src="docs/img/img_library_eng.png" alt="Library scheme" height=""/>\n\n\n## Installation\n\n! Note: The functionality of topic models training is available only for linux distributions.\n\n```pip install -r requirements.txt```  \n\n```python -m spacy download en_core_web_sm```\n\n```export PYTHONPATH="${PYTHONPATH}:/path/to/src"```\n\n[//]: # (## Dataset and )\n\n## Quickstart\n\nThe notebook with an example is available in ```examples``` folder.\n\n## Distributed version\n\nDistributed version to run experiments on kubernetes is available in ```autotm-distributed``` brunch. Still this version is in development stage and will be transfered to separate repository.\n\n## Backlog:\n- [ ] Add tests\n- [ ] Add new multi-stage \n \n## Citation\n\n```bibtex\n@article{10.1093/jigpal/jzac019,\n    author = {Khodorchenko, Maria and Butakov, Nikolay and Sokhin, Timur and Teryoshkin, Sergey},\n    title = "{ Surrogate-based optimization of learning strategies for additively regularized topic models}",\n    journal = {Logic Journal of the IGPL},\n    year = {2022},\n    month = {02},\n    issn = {1367-0751},\n    doi = {10.1093/jigpal/jzac019},\n    url = {https://doi.org/10.1093/jigpal/jzac019},}\n\n```\n',
    'author': 'Khodorchenko Maria',
    'author_email': 'mariyaxod@yandex.ru',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://autotm.readthedocs.io/en/latest/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.10',
}


setup(**setup_kwargs)



# source: https://github.com/machine-intelligence-laboratory/OptimalNumberOfTopics/blob/master/topnum/scores/intratext_coherence_score.py

import dill
import numpy as np
import pandas as pd
import sys
import tqdm
import warnings

from collections import defaultdict
from enum import Enum, IntEnum, auto
from typing import (
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
    Union,
)

from base_score import BaseScore

VW_TEXT_COL = "vw_text"
RAW_TEXT_COL = "raw_text"


class TextType(Enum):
    VW_TEXT = VW_TEXT_COL
    RAW_TEXT = RAW_TEXT_COL


class ComputationMethod(IntEnum):
    """
    Ways to compute intra-text coherence
    (see more about coherence below in IntratextCoherenceScore)
    Attributes
    ----------
    SEGMENT_LENGTH :
        Estimate the length of topic segments
    SEGMENT_WEIGHT :
        Estimate the weight of topic segment
        (weight - sum of specificities for the topic over words in segment)
    SUM_OVER_WINDOW :
        Sum of specificities for the topic over words in given window.
        The process is as follows:
        word of the topic is found in text, it is the center of the first window;
        next word of the topic is found (outside of the previous window), window; etc
    """

    SEGMENT_LENGTH = auto()
    SEGMENT_WEIGHT = auto()
    SUM_OVER_WINDOW = auto()


class WordTopicRelatednessType(IntEnum):
    """
    Word-topic relatedness estimate
    Attributes
    ----------
    PWT :
        p(w | t)
    PTW :
        p(t | w)
    """

    PWT = auto()
    PTW = auto()


class SpecificityEstimationMethod(IntEnum):
    """
    Way to estimate how particular word is specific for particular topic.
    Unlike probability, eg. p(w | t), specificity_estimation takes into account
    values for all topics, eg. p(w | t_1), p(w | t_2), ..., p(w | t_n):
    the higher the value p(w | t) comparing other p(w | t_i),
    the higher the specificity_estimation of word "w" for the topic "t"
    Attributes
    ----------
    NONE :
        Don't try to estimate specificity_estimation, return the probability as is
    MAXIMUM :
        From probability, corresponding to word and topic,
        extract *maximum* among probabilities for the word and other topics
    AVERAGE :
        From probability, corresponding to word and topic,
        extract *average* among probabilities for the word and other topics
    """

    NONE = auto()
    MAXIMUM = auto()
    AVERAGE = auto()


class IntratextCoherenceScore(BaseScore):
    """Computes intratext coherence
    For each topic of topic model its distribution throughout document collection is observed.
    Hypothetically, the better the topic, the more often it is represented by
    long segments of words highly related to the topic.
    The score tries to bring to life this idea.
    For more details one may see the article http://www.dialog-21.ru/media/4281/alekseevva.pdf
    """

    def __init__(  # noqa: C901
        self,
        dataset: Union[Dataset, str],
        name: str = None,
        should_compute: Callable[[int], bool] = None,
        keep_dataset_in_memory: bool = None,
        keep_dataset: bool = True,
        documents: List[str] = None,
        documents_fraction: float = 1.0,
        text_type: TextType = TextType.VW_TEXT,
        computation_method: ComputationMethod = ComputationMethod.SEGMENT_WEIGHT,
        word_topic_relatedness: WordTopicRelatednessType = WordTopicRelatednessType.PWT,
        specificity_estimation: SpecificityEstimationMethod = SpecificityEstimationMethod.NONE,
        max_num_out_of_topic_words: int = 10,
        window: int = 20,
        start_fit_iteration: int = 0,
        fit_iteration_step: int = 1,
        seed: int = 11221963,
        verbose: bool = False,
    ):
        """
        Parameters
        ----------
        name:
            Name of the score
        dataset : Dataset
            Dataset with document collection, or path to dataset
            (any model passed to `call()` is supposed to be trained on it)
        keep_dataset_in_memory
            Whether to keep `dataset` in memory or not
            (parameter `_small_data` of the `dataset` object).
            If `dataset` is given as object of type `Dataset` (and not as `str` path to dataset),
            the parameter will be set equal to `dataset._small_data`.
            Otherwise, the default value is `True` and `dataset._small_data` will be overwritten.
        keep_dataset
            Whether to keep `dataset` constantly as inner part of the score,
            or recreate it for each `call()` invocation and then dispose
        documents : list of str
            Which documents from the dataset are to be used for computing coherence
        documents_fraction
            The fraction of all the documents in the Dataset to be used for coherence computation
            if `documents` parameter is not specified
        text_type : TextType
            What text to use when computing coherence: raw text or VW text
            Preferable to use VW (as it is usually preprocessed, stop-words removed etc.),
            and with words in *natural order*.
            Score needs "real" text to compute coherence
        computation_method : ComputationMethod
            The way to compute intra-text coherence
        word_topic_relatedness : WordTopicRelatednessType
            How to estimate word relevance to topic: using p(w | t) or p(t | w)
        specificity_estimation : SpecificityEstimationMethod
            How to estimate specificity of word to topic
        max_num_out_of_topic_words : int
            In case computation_method = ComputationMethod.SEGMENT_LENGTH or
            ComputationMethod.SEGMENT_WEIGHT:
            Maximum number of words not of the topic which can be encountered without stopping
            the process of adding words to the current segment
        window : int
            In case computation_method = ComputationMethod.SUM_OVER_WINDOW:
            Window width. So the window will be the words with positions
            in [current position - window / 2, current position + window / 2)
        start_fit_iteration
            Indicates how many calls are skipped before the actual score is calculated.
            Replaces not calculated values with placeholders
            (for consistency of score values with number of model fit iterations).
        fit_iteration_step
            Number of iterations between `score.call()` invocations which actually update the score
        seed
            Random seed used for documents subsampling if `documents` parameter is not specified
        Notes
        -----
        Parameters `start_fit_iteration` and `fit_iteration_step` are introduced
        to reduce the time needed for one model training.
        If one is interested only in the last score value
        at the end of the training process (and not in the dependence of score on iteration),
        one should adjust `start_fit_iteration` and `fit_iteration_step` correspondingly.
        For example:
        >>> # dataset = Dataset(...)
        >>> # topic_model = TopicModel(...)
        >>> num_iterations = 100
        >>> topic_model.custom_scores['intratext_coherence'] = IntratextCoherenceScore(
        >>>     dataset,
        >>>     start_fit_iteration=num_iterations - 1  # last iteration: starting from zero
        >>> )
        >>> topic_model._fit(dataset.get_batch_vectorizer(), num_iterations=num_iterations)
        """
        # TODO: word_topic_relatedness seems to be connected with TopTokensViewer stuff
        super().__init__(name=name, should_compute=should_compute)

        self._keep_dataset = keep_dataset

        if isinstance(dataset, str):
            if keep_dataset_in_memory is None:
                keep_dataset_in_memory = True

            dataset = Dataset(data_path=dataset, keep_in_memory=keep_dataset_in_memory)

        self._keep_dataset_in_memory = dataset._small_data

        if not isinstance(dataset, Dataset):
            raise TypeError(
                f'Got "{type(dataset)}" as "dataset". Expect it to derive from "Dataset"'
            )

        if not isinstance(text_type, TextType):
            raise TypeError(
                f'Wrong "text_type": "{text_type}". ' f'Expect to be "{TextType}"'
            )

        if not isinstance(computation_method, ComputationMethod):
            raise TypeError(
                f'Wrong "computation_method": "{computation_method}". '
                f'Expect to be "{ComputationMethod}"'
            )

        if not isinstance(word_topic_relatedness, WordTopicRelatednessType):
            raise TypeError(
                f'Wrong "word_topic_relatedness": "{word_topic_relatedness}". '
                f'Expect to be "{WordTopicRelatednessType}"'
            )

        if not isinstance(specificity_estimation, SpecificityEstimationMethod):
            raise TypeError(
                f'Wrong "specificity_estimation": "{specificity_estimation}". '
                f'Expect to be "{SpecificityEstimationMethod}"'
            )

        if not isinstance(max_num_out_of_topic_words, int):
            raise TypeError(
                f'Wrong "max_num_out_of_topic_words": "{max_num_out_of_topic_words}". '
                f'Expect to be "int"'
            )

        if not isinstance(window, int):
            raise TypeError(f'Wrong "window": "{window}". Expect to be "int"')

        if window < 0 or (
            window == 0 and computation_method == ComputationMethod.SUM_OVER_WINDOW
        ):
            raise ValueError(
                f'Wrong value for "window": "{window}". '
                f"Expect to be non-negative. And greater than zero in case "
                f"computation_method == ComputationMethod.SUM_OVER_WINDOW"
            )

        if not isinstance(start_fit_iteration, int):
            raise TypeError(
                f'Wrong "start_fit_iteration": "{start_fit_iteration}".'
                f' Expect to be "int"'
            )

        if not isinstance(fit_iteration_step, int):
            raise TypeError(
                f'Wrong "fit_iteration_step": "{start_fit_iteration}".'
                f' Expect to be "int"'
            )
        if fit_iteration_step <= 0:
            raise ValueError(
                f'Wrong "fit_iteration_step": "{fit_iteration_step}".'
                f" Expect to be > 0"
            )

        if documents_fraction <= 0:
            raise ValueError(
                f'Wrong "documents_fraction": "{documents_fraction}".'
                f" Expect to be in (0, 1]"
            )
        if documents_fraction > 1.0:
            warnings.warn(
                f"Parameter documents_fraction={documents_fraction} can't be bigger than 1.0"
                f" Setting it equal to 1.0"
            )

            documents_fraction = 1.0

        self._dataset = dataset
        self._dataset_file_path = dataset._data_path
        self._dataset_internals_folder_path = dataset._internals_folder_path

        self._text_type = text_type
        self._computation_method = computation_method
        self._word_topic_relatedness = word_topic_relatedness
        self._specificity_estimation_method = specificity_estimation
        self._max_num_out_of_topic_words = max_num_out_of_topic_words
        self._window = window

        self._verbose = verbose

        self._current_iteration = 0
        self._start_fit_iteration = start_fit_iteration
        self._fit_iteration_step = fit_iteration_step

        if documents is not None:
            self._documents = documents
        else:
            all_documents = list(self._dataset.get_dataset().index)
            documents_fraction = min(documents_fraction, 1.0)
            num_documents_to_choose = int(
                np.ceil(len(all_documents) * documents_fraction)
            )
            custom_random = np.random.RandomState(seed)

            self._documents = list(
                custom_random.choice(
                    all_documents, size=num_documents_to_choose, replace=False
                )
            )

    def __repr__(self):
        return (
            f"{self.__class__.__name__}("
            f"text_type={self._text_type!r}"
            f"computation_method={self._computation_method!r}"
            f"word_topic_relatedness={self._word_topic_relatedness!r}"
            f"specificity_estimation_method={self._specificity_estimation_method!r}"
            f"max_num_out_of_topic_words={self._max_num_out_of_topic_words!r}"
            f"window={self._window!r}"
            f")"
        )

    @property
    def dataset(self) -> Dataset:
        return self._dataset

    @dataset.setter
    def dataset(self, new_dataset: Dataset) -> None:
        self._dataset = new_dataset
        self._dataset_file_path = new_dataset._data_path
        self._dataset_internals_folder_path = new_dataset._internals_folder_path
        self._keep_dataset_in_memory = new_dataset._small_data

    def save(self, path: str) -> None:
        dataset = self._dataset
        self._dataset = None

        with open(path, "wb") as f:
            dill.dump(self, f)

        self._dataset = dataset

    @classmethod
    def load(cls, path: str):
        """
        Parameters
        ----------
        path
        Returns
        -------
        IntratextCoherenceScore
        """
        score: IntratextCoherenceScore

        with open(path, "rb") as f:
            score = dill.load(f)

        if not score._keep_dataset:
            score._dataset = None
        else:
            score._dataset = Dataset(
                score._dataset_file_path,
                internals_folder_path=score._dataset_internals_folder_path,
                keep_in_memory=score._keep_dataset_in_memory,
            )

        return score

    def call(self, model: BaseModel, **kwargs) -> float:
        if (
            self._current_iteration - self._start_fit_iteration
        ) % self._fit_iteration_step != 0:
            self._current_iteration += 1

            return float("nan")

        try:
            if self._dataset is None:
                self._dataset = Dataset(
                    self._dataset_file_path,
                    internals_folder_path=self._dataset_internals_folder_path,
                    keep_in_memory=self._keep_dataset_in_memory,
                )

            topic_coherences = self.compute(model, None)

            coherence_values = list(
                v if v is not None else 0.0  # TODO: state the behavior clearer somehow
                for v in topic_coherences.values()
            )

            self._current_iteration += 1

            return float(np.median(coherence_values))  # TODO: or mean?

        finally:
            if not self._keep_dataset:
                self._dataset = None

    def compute(
        self, model: BaseModel, topics: List[str] = None, documents: List[str] = None
    ) -> Dict[str, Optional[float]]:
        if not isinstance(model, BaseModel):
            raise TypeError(
                f'Got "{type(model)}" as "model". '
                f'Expect it to derive from "BaseModel"'
            )

        if topics is None:
            topics = IntratextCoherenceScore._get_topics(model)

        if documents is None:
            documents = list(self._documents)

        if not isinstance(topics, list):
            raise TypeError(
                f'Got "{type(topics)}" as "topics". Expect list of topic names'
            )

        if not isinstance(documents, list):
            raise TypeError(
                f'Got "{type(documents)}" as "documents". Expect list of document ids'
            )

        word_topic_relatednesses = self._get_word_topic_relatednesses(model)

        topic_document_coherences = np.zeros((len(topics), len(documents)))
        document_indices_with_topic_coherence = defaultdict(list)

        if not self._verbose:
            document_enumeration = enumerate(documents)
        else:
            document_enumeration = tqdm.tqdm(
                enumerate(documents), total=len(documents), file=sys.stdout
            )

        for document_index, document in document_enumeration:
            for topic_index, topic in enumerate(topics):
                # TODO: read document text only once for all topics
                topic_coherence = self._compute_coherence(
                    topic, document, word_topic_relatednesses
                )

                if topic_coherence is not None:
                    topic_document_coherences[
                        topic_index, document_index
                    ] = topic_coherence
                    document_indices_with_topic_coherence[topic].append(document_index)

        topic_coherences = [
            topic_document_coherences[
                topic_index, document_indices_with_topic_coherence[topic]
            ]
            if len(document_indices_with_topic_coherence) > 0
            else list()
            for topic_index, topic in enumerate(topics)
        ]

        return dict(
            zip(
                topics,
                [
                    float(np.mean(coherence_values))
                    if len(coherence_values) > 0
                    else None
                    for coherence_values in topic_coherences
                ],
            )
        )

    @staticmethod
    def _get_topics(model):
        return list(model.get_phi().columns)

    def _get_word_topic_relatednesses(self, model) -> pd.DataFrame:
        phi = model.get_phi()

        word_topic_probs = self._get_word_topic_probs(phi)

        if self._specificity_estimation_method == SpecificityEstimationMethod.NONE:
            pass

        elif self._specificity_estimation_method == SpecificityEstimationMethod.AVERAGE:
            word_topic_probs[:] = word_topic_probs.values - np.sum(
                word_topic_probs.values, axis=1, keepdims=True
            ) / max(  # noqa E131
                word_topic_probs.shape[1], 1
            )  # noqa E131

        elif self._specificity_estimation_method == SpecificityEstimationMethod.MAXIMUM:
            new_columns = []

            for t in word_topic_probs.columns:
                new_column = word_topic_probs[t].values - np.max(
                    word_topic_probs[word_topic_probs.columns.difference([t])].values,
                    axis=1,
                )
                new_columns.append(list(new_column))

            word_topic_probs[:] = np.array(new_columns).T

        return word_topic_probs

    def _get_word_topic_probs(self, phi: pd.DataFrame) -> pd.DataFrame:
        if self._word_topic_relatedness == WordTopicRelatednessType.PWT:
            return phi

        elif self._word_topic_relatedness == WordTopicRelatednessType.PTW:
            # Treat all topics as equally probable
            eps = np.finfo(float).tiny

            pwt = phi
            pwt_values = pwt.values

            return pd.DataFrame(
                index=pwt.index,
                columns=pwt.columns,
                data=pwt_values / (pwt_values.sum(axis=1).reshape(-1, 1) + eps),
            )

        assert False

    def _compute_coherence(self, topic, document, word_topic_relatednesses):
        assert isinstance(self._computation_method, ComputationMethod)

        words = self._get_words(document)

        if self._computation_method == ComputationMethod.SUM_OVER_WINDOW:
            average_sum_over_window = self._sum_relatednesses_over_window(
                topic, words, word_topic_relatednesses
            )

            return average_sum_over_window

        (
            topic_segment_length,
            topic_segment_weight,
        ) = self._compute_segment_characteristics(
            topic, words, word_topic_relatednesses
        )

        if self._computation_method == ComputationMethod.SEGMENT_LENGTH:
            return topic_segment_length

        elif self._computation_method == ComputationMethod.SEGMENT_WEIGHT:
            return topic_segment_weight

    def _get_words(self, document):
        def get_biggest_modality_or_default():
            modalities = list(self._dataset.get_possible_modalities())

            if len(modalities) == 0:
                return DEFAULT_ARTM_MODALITY

            modalities_vocabulary_sizes = list(
                map(lambda m: self._dataset.get_dataset().loc[m].shape[0], modalities)
            )

            return modalities[np.argmax(modalities_vocabulary_sizes)]

        if self._text_type == TextType.RAW_TEXT:
            text = self._dataset.get_source_document(document).values[
                0, 0
            ]  # TODO: this way?
            modality = get_biggest_modality_or_default()

            return list(map(lambda w: (modality, w), text.split()))

        if self._text_type == TextType.VW_TEXT:
            text = self._dataset.get_vw_document(document).values[
                0, 0
            ]  # TODO: this way?

            words = []
            modality = None

            # TODO: there was similar bunch of code somewhere...
            for word in text.split()[1:]:  # skip document id
                if word.startswith(MODALITY_START_SYMBOL):
                    modality = word[1:]

                    continue

                word = word.split(":")[0]

                if modality is not None:
                    word = (modality, word)  # phi multiIndex
                else:
                    word = (DEFAULT_ARTM_MODALITY, word)

                words.append(word)

            return words

        assert False

    def _compute_segment_characteristics(
        self, topic, words, word_topic_relatednesses: pd.DataFrame
    ) -> Tuple[float, float]:
        topic_segment_lengths = []
        topic_segment_weights = []

        topic_index = word_topic_relatednesses.columns.get_loc(topic)
        word_topic_indices = np.argmax(word_topic_relatednesses.values, axis=1)

        def get_word_topic_index(word):
            if word not in word_topic_relatednesses.index:
                return -1
            else:
                return word_topic_indices[word_topic_relatednesses.index.get_loc(word)]

        index = 0

        while index < len(words):
            original_index = index

            if get_word_topic_index(words[index]) != topic_index:
                index += 1

                continue

            segment_length = 1
            segment_weight = IntratextCoherenceScore._get_relatedness(
                words[index], topic, word_topic_relatednesses
            )

            num_out_of_topic_words = 0

            index += 1

            while (
                index < len(words)
                and num_out_of_topic_words < self._max_num_out_of_topic_words
            ):
                if get_word_topic_index(words[index]) != topic_index:
                    num_out_of_topic_words += 1
                else:
                    segment_length += 1
                    segment_weight += IntratextCoherenceScore._get_relatedness(
                        words[index], topic, word_topic_relatednesses
                    )

                    num_out_of_topic_words = 0

                index += 1

            topic_segment_lengths.append(segment_length)
            topic_segment_weights.append(segment_weight)

            assert index > original_index

        if len(topic_segment_lengths) == 0:
            return None, None
        else:
            return np.mean(topic_segment_lengths), np.mean(topic_segment_weights)

    def _sum_relatednesses_over_window(
        self, topic, words, word_topic_relatednesses
    ) -> float:
        topic_index = word_topic_relatednesses.columns.get_loc(topic)
        word_topic_indices = np.argmax(word_topic_relatednesses.values, axis=1)

        def get_word_topic_index(word):
            if word not in word_topic_relatednesses.index:
                return -1
            else:
                return word_topic_indices[word_topic_relatednesses.index.get_loc(word)]

        def find_next_topic_word(starting_index: int) -> int:
            index = starting_index

            while (
                index < len(words) and get_word_topic_index(words[index]) != topic_index
            ):
                index += 1

            if index == len(words):
                return -1  # failed to find next topic word

            return index

        word_index = find_next_topic_word(0)

        if word_index == -1:
            return None

        sums = list()

        while word_index < len(words) and word_index != -1:
            original_word_index = word_index

            window_lower_bound = word_index - int(np.floor(self._window // 2))
            window_upper_bound = word_index + int(np.ceil(self._window // 2))

            sum_in_window = np.sum(
                [
                    IntratextCoherenceScore._get_relatedness(
                        w, topic, word_topic_relatednesses
                    )
                    for w in words[window_lower_bound:window_upper_bound]
                ]
            )

            sums.append(sum_in_window)

            word_index = find_next_topic_word(window_upper_bound)

            assert word_index > original_word_index or word_index == -1

        return np.mean(sums)

    @staticmethod
    def _get_relatedness(word, topic, word_topic_relatednesses: pd.DataFrame) -> float:
        if word in word_topic_relatednesses.index:
            return word_topic_relatednesses.loc[word, topic]

        # TODO: throw Warning or log somewhere?
        return np.mean(word_topic_relatednesses.values)
